"""
    File: writers.py
    Author: Wes Holliday (wesholliday@berkeley.edu) and Eric Pacuit (epacuit@umd.edu)
    Date: March 17, 2024

    Functions to write election data to a file.
"""

from pref_voting.rankings import Ranking
from pref_voting.profiles import Profile
from pref_voting.profiles_with_ties import ProfileWithTies
from preflibtools.instances import OrdinalInstance
import csv
import json
import numpy as np

def to_preflib_instance(profile):
    """
    Returns an instance of the ``OrdinalInstance`` class from the ``preflibtools`` package (see https://preflib.github.io/preflibtools/usage.html#ordinal-preferences).  

    Args:
        profile: A Profile or ProfileWithTies object.

    Returns:
        An instance of the ``OrdinalInstance`` class from preflibtools.
        
    """
    assert type(profile) in [Profile, ProfileWithTies], "Must be a Profile or ProfileWithTies object to convert to a preflib OrdinalInstance."

    instance = OrdinalInstance()
    vote_map = dict()
    cand_to_cidx = {c: i for i, c in enumerate(profile.candidates)}
    cmap = {i: profile.cmap[c] for c, i in cand_to_cidx.items()}
    for r,c in zip(*profile.rankings_counts):
        ranking = tuple([tuple([cand_to_cidx[_c] for _c in indiff]) for indiff in r.to_indiff_list()]) if type(r) == Ranking else tuple([(c,) for c in r])
        if ranking in vote_map.keys():
            vote_map[ranking] += c
        else:
            vote_map[ranking] = c
    instance.append_vote_map(vote_map)  
    instance.alternatives_name = cmap  
    return instance   

def write_preflib(profile, filename):
    """
    Write a profile to a file in the PrefLib format.

    Args:
        profile: A Profile or ProfileWithTies object.
        filename: The name of the file to write the profile to.

    Returns:
        The name of the file the profile was written to.
    """
    assert type(profile) in [Profile, ProfileWithTies], "Must be a Profile or ProfileWithTies object to write in the preflib format."

    instance = to_preflib_instance(profile)
    preflib_type = instance.infer_type()
    instance.write(filename)

    if not filename.endswith(preflib_type):
        filename += f".{preflib_type}"

    print(f"Election written to {filename}.")
    return f"{filename}"

def write_csv(profile, filename, csv_format="candidate_columns"):
    """
    Write a profile to a file in CSV format.

    Args:
        profile: A Profile or ProfileWithTies object.
        filename: The name of the file to write the profile to.
        csv_format: The format to write the profile in.  Defaults to "candidate_columns".  The other option is "rank_columns".
    """
    assert type(profile) in [Profile, ProfileWithTies], "Must be a Profile or ProfileWithTies object to write in the csv format."

    candidates = profile.candidates

    if not filename.endswith(".csv"):
        filename += ".csv"

    if csv_format == "rank_columns":

        assert profile.is_truncated_linear, "The profile must be truncated linear to use the rank_columns csv format."

        ranks = range(1, len(candidates) + 1)
        with open(filename, mode='w') as file:
            writer = csv.writer(file)
            writer.writerow([f"Rank{_r}" for _r in ranks])
            for indiff_list in profile.rankings_as_indifference_list:
                ranking = [str(profile.cmap[cs[0]]) for cs in indiff_list]
                writer.writerow(ranking if len(ranking) == len(candidates) else ranking + ["skipped"] * (len(candidates) - len(ranking)))
        
        print(f"Election written to {filename}.")

        return filename

    elif csv_format == "candidate_columns":
        
        prof = profile.to_profile_with_ties() if type(profile) == Profile else profile

        rs, cs = prof.rankings_counts
        anon_rankings = []
        for r, count in zip(rs, cs): 
            r.normalize_ranks()
            found_it = False
            for r_c in anon_rankings: 
                if r_c[0] == r: 
                    found_it = True
                    r_c[1] += count
            if not found_it: 
                anon_rankings.append([r, count])
            
        with open(filename, mode='w') as file:
            writer = csv.writer(file)
            writer.writerow([profile.cmap[c] for c in candidates] + ["#"])
            for r,count in anon_rankings:
                writer.writerow([r.rmap[c] if r.is_ranked(c) else "" for c in candidates] + [count])

        print(f"Election written to {filename}.")

        return filename

def write_json(profile, filename):
    """
    Write a profile to a file in JSON format.

    Args:
        profile: A Profile or ProfileWithTies object.
        filename: The name of the file to write the profile to.
    
    Returns:
        The name of the file the profile was written to.
    """
    assert type(profile) in [Profile, ProfileWithTies], "Cannot write to the abif format."

    if not filename.endswith(".json"):
        filename += ".json"

    prof = profile.to_profile_with_ties() if type(profile) == Profile else profile

    prof_as_dict = {
            "candidates": profile.candidates,
            "rankings": [{"ranking": {
                int(cand) if isinstance(cand, np.int64) else cand: int(rank) 
                for cand,rank in r.rmap.items()}, 
                "count": int(c)} 
                for r,c in zip(*prof.rankings_counts)],
            "cmap": profile.cmap
        }
    with open(filename, "w") as f:
        json.dump(prof_as_dict, f)

    print(f"Election written to {filename}.")
    return filename

def write_abif(profile, filename):
    """
    Write a profile to a file in ABIF format.
    
    The ABIF format is explained here: https://electowiki.org/wiki/ABIF.

    Args:
        profile: A Profile or ProfileWithTies object.
        filename: The name of the file to write the profile to.
    
    Returns:
        The name of the file the profile was written to.
    """

    assert type(profile) in [Profile, ProfileWithTies], "Cannot write to the abif format."

    if not filename.endswith(".abif"):
        filename += ".abif"
        
    with open(filename, mode='w') as file:
        file.write(f"# {len(profile.candidates)} candidates\n")
        for c in profile.candidates:
            file.write(f"={c} : [{profile.cmap[c]}]\n")
        for r, c in zip(*profile.rankings_counts):
            indiff_list = r.to_indiff_list() if type(r) == Ranking else [(c,) for c in r]
            file.write(f"{c}:{'>'.join(['='.join([str(c) for c in cs]) for cs in indiff_list])}\n")
                    
        print(f"Election written to {filename}.")

        return filename

def write_grade_profile_to_abif(profile):
    """
    Write a profile to a file in ABIF format.

    Args:
        profile: A Profile object.
    """
    pass

def write_spatial_profile_to_json(spatial_profile, filename):
    """
    Write a spatial profile to a file in JSON format.

    Args:
        spatial_profile: A SpatialProfile object.

    Returns:
        The name of the file the spatial profile was written to.
    """
     
    if not filename.endswith(".json"):
        filename += ".json"

    with open(filename, "w") as f:
        spatial_profile_dict = {
            "cand_names": spatial_profile.candidates,
            "voter_names": spatial_profile.voters,  
            "candidates": {c: list(spatial_profile.candidate_position(c)) for c in spatial_profile.candidates},
            "voters": {v: list(spatial_profile.voter_position(v)) for v in spatial_profile.voters}
        }
        json.dump(spatial_profile_dict, f)

    print(f"Spatial profile written to {filename}.")

    return filename

def write(
        edata, 
        filename, 
        file_format='preflib', 
        csv_format="candidate_columns"):
    """
    Write election data to ``filename`` in the format specified in ``file_format``.

    Args:
        edata: Election data to write.
        filename: The name of the file to write the election data to.
        file_format: The format to write the election data in. Defaults to "preflib".  The other options are "csv", "json", and "abif".
        csv_format: The format to write the election data in if the file format is "csv".  Defaults to 'candidate_columns'.  The other option is ``rank_columns``.

    Returns:
        The name of the file the election data was written to.
       
    Note: 
        There are two formats for the csv file: "rank_columns" and "candidate_columns".  The "rank_columns" format is used when the csv file contains a column for each rank and the rows are the candidates at that rank (or "skipped" if the ranked is skipped).  The "candidate_columns" format is used when the csv file contains a column for each candidate and the rows are the rank of the candidates (or the empty string if the candidate is not ranked).

    """

    if file_format == 'preflib':
        return write_preflib(edata, filename)
    elif file_format == 'csv':
        return write_csv(edata, filename, csv_format=csv_format)
    elif file_format == 'json':
        return write_json(edata, filename)
    elif file_format == 'abif':
        return write_abif(edata, filename)
    else:
        raise ValueError(f"File format {file_format} not recognized.")


'''
    File: voting_methods.py
    Author: Wes Holliday (wesholliday@berkeley.edu) and Eric Pacuit (epacuit@umd.edu)
    Date: Nove 21, 2024
    
    Implementations of probabilistic voting methods.
'''

from pref_voting.prob_voting_method import  *
from pref_voting.weighted_majority_graphs import  MajorityGraph, MarginGraph
from scipy.optimize import linprog

import random
import nashpy as nash

@pvm(name="Random Dictator")
def random_dictator(profile, curr_cands = None): 
    '''Returns lottery over the candidates that is proportional to the Plurality scores. 

    Args:
        profile (Profile): A Profile object.
        curr_cands (list): A list of candidates to restrict the ranking to. If ``None``, then the ranking is over the entire domain of the profile.

    Returns:
        dict: A dictionary mapping candidates to probabilities.
    ''' 
    
    plurality_scores = profile.plurality_scores(curr_cands = curr_cands)
    total_plurality_scores = sum(list(plurality_scores.values()))

    return {c: plurality_scores[c] / total_plurality_scores for c in plurality_scores.keys()}

@pvm(name="Random Dictator on the Beta-Uncovered Set")
def RaDiUS(profile, curr_cands = None, beta = 0.5):
    """
    Runs the Random Dictator method on the profile restricted to the beta-uncovered set, as proposed by Charikar et al. (https://arxiv.org/abs/2306.17838).

    Args:
        profile (Profile): An anonymous profile of linear orders
        curr_cands (List[int], optional): Candidates to consider. Defaults to all candidates if not provided. 

    Returns:
        dict: Maps each candidate to their probability of winning under the RaDiUS method.

    """
    from pref_voting.margin_based_methods import beta_uncovered_set

    curr_cands = profile.candidates if curr_cands is None else curr_cands

    rd_dist = random_dictator(profile, curr_cands = beta_uncovered_set(profile, curr_cands = curr_cands, beta = beta))
    
    rd_dist.update({c:0 for c in curr_cands if c not in rd_dist.keys()})

    return rd_dist

@pvm(name="Proportional Borda")
def pr_borda(profile, curr_cands=None): 
    '''Returns lottery over the candidates that is proportional to the Borda scores.
    
    Args:   
        profile (Profile): A Profile object.
        curr_cands (list): A list of candidates to restrict the ranking to. If ``None``, then the ranking is over the entire domain of the profile.

    Returns:
        dict: A dictionary mapping candidates to probabilities.
    
    '''
    borda_scores = profile.borda_scores(curr_cands=curr_cands)
    total_borda_scores = sum(list(borda_scores.values()))

    return {c: borda_scores[c] / total_borda_scores for c in borda_scores.keys()}

def clean_and_normalize(probs, threshold=1e-10):
    # Set  negative or small positive values to zero
    probs = np.where(probs < threshold, 0, probs)
    
    # Renormalize to ensure the probabilities sum to 1
    total = np.sum(probs)
    if total > 0:
        probs /= total
    return probs

def _maximal_lottery_enumeration(edata, curr_cands=None, margin_transformation=lambda x: x):
    '''
    Implementation of maximal lotteries. See http://dss.in.tum.de/files/brandt-research/fishburn_slides.pdf 
    
    Returns a randomly chosen maximal lottery.
    '''

    candidates = edata.candidates if curr_cands is None else curr_cands
    m_matrix, cand_to_cidx = edata.strength_matrix(curr_cands=candidates)

    A = np.array([[margin_transformation(m) for m in row] for row in m_matrix])

    # Create the game
    game = nash.Game(A)
    equilibria = []
    try:
        equilibria = list(game.vertex_enumeration())
        #print("Vertex Enumeration found equilibria.")
    except Exception as e:
        print(f"Vertex Enumeration failed: {e}")

    # Backup method 1: Support Enumeration
    if not equilibria:
        try:
            equilibria = list(game.support_enumeration())
            #print("Support Enumeration found equilibria.")
        except Exception as e:
            print(f"Support Enumeration failed: {e}")

    if len(equilibria) == 0:
        return {c: 1 / len(candidates) for c in candidates}
    else:
        # average the  equilibria component-wise to get a single equilibrium
        eq_probs = [np.mean([eq[0][idx] for eq in equilibria]) 
                    for idx in range(len(candidates))]

        eq_probs = clean_and_normalize(np.array(eq_probs))
        
        # Return the result as a dictionary
        return {c: eq_probs[cand_to_cidx(c)] for c in candidates}


def _maximal_lottery_lp(edata, curr_cands=None, margin_transformation=lambda x: x):
    '''
    Implementation of maximal lotteries using linear programming.
    '''
    candidates = edata.candidates if curr_cands is None else curr_cands
    m_matrix, cand_to_cidx = edata.strength_matrix(curr_cands=candidates)

    A = np.array([[margin_transformation(m) for m in row] for row in m_matrix])

    num_cands = len(candidates)
    c = np.zeros(num_cands + 1)
    c[-1] = -1  # Coefficient for v in the objective function (maximize v)

    # Inequalities: A^T p - v * 1 >= 0 => A^T p - v * 1 - s = 0 (s >= 0)
    # We need to convert this to the form: A_ub x <= b_ub

    A_ub = np.hstack([-A.T, np.ones((num_cands, 1))])
    b_ub = np.zeros(num_cands)

    # Equalities: sum p_i = 1
    A_eq = np.zeros((1, num_cands + 1))
    A_eq[0, :num_cands] = 1
    b_eq = np.array([1])

    bounds = [(0, None)] * num_cands + [(None, None)]  # p_i >= 0, v free

    res = linprog(c, A_ub=A_ub, b_ub=b_ub, A_eq=A_eq, b_eq=b_eq, bounds=bounds, method='highs')

    if res.success:
        eq_probs = res.x[:num_cands]
        # Normalize to account for numerical errors
        eq_probs = np.maximum(eq_probs, 0)
        eq_probs /= np.sum(eq_probs)
        return {c: eq_probs[cand_to_cidx(c)] for c in candidates}
    else:
        # If LP fails, default to uniform distribution
        return {c: 1 / len(candidates) for c in candidates}


@pvm(name="C1 Maximal Lottery")
def c1_maximal_lottery(edata, curr_cands=None, algorithm='enumeration'): 

    '''Returns the C1 maximal lottery over the candidates.  See http://dss.in.tum.de/files/brandt-research/fishburn_slides.pdf.
    
    Args:   
        edata (Profile, MarginGraph): A Profile object.
        curr_cands (list): A list of candidates to restrict the ranking to. If ``None``, then the ranking is over the entire domain of the profile.
        algorithm (str): The algorithm to use. Either 'enumeration' or 'lp'. Defaults to 'enumeration'.
    
    Returns:
        dict: A dictionary mapping candidates to probabilities.

    .. note::
        The 'enumeration' algorithm averages over the extremal maximal lotteries.   The 'lp' is faster, but only returns a single maximal lottery (not necessarily the average)

    '''

    if type(edata) == MajorityGraph:
        # if edata is a MajorityGraph, we need to add margins for the following code to work.  The margins do not matter when finding the c1 maximal lottery.

        candidates = edata.candidates if curr_cands is None else curr_cands 
          
        edata = MarginGraph(candidates, [(c1, c2, 1) for c1, c2 in edata.edges if (c1 in candidates and c2 in candidates)])

    if algorithm == 'enumeration':
        return _maximal_lottery_enumeration(edata, curr_cands=curr_cands, margin_transformation = np.sign)
        
    elif algorithm == 'lp':
        return _maximal_lottery_lp(edata, curr_cands=curr_cands, margin_transformation = np.sign)
    else:
        raise ValueError(f"Unknown algorithm: {algorithm}")

@pvm(name="Maximal Lottery")
def maximal_lottery(edata, curr_cands=None, algorithm='lp'): 
    '''Returns the maximal lottery over the candidates.  See http://dss.in.tum.de/files/brandt-research/fishburn_slides.pdf.
    
    Args:   
        edata (Profile, MarginGraph): A Profile object.
        curr_cands (list): A list of candidates to restrict the ranking to. If ``None``, then the ranking is over the entire domain of the profile.
        algorithm (str): The algorithm to use. Either 'enumeration' or 'lp'. Defaults to 'enumeration'.

    Returns:
        dict: A dictionary mapping candidates to probabilities.
    
    .. note::
        The 'enumeration' algorithm averages over the extremal maximal lotteries.   The 'lp' is faster, but only returns a single maximal lottery (not necessarily the average)
    
    
    '''

    if algorithm == 'enumeration':
        return _maximal_lottery_enumeration(edata, curr_cands=curr_cands, margin_transformation = lambda x: x)
    
    elif algorithm == 'lp':
        return _maximal_lottery_lp(edata, curr_cands=curr_cands, margin_transformation = lambda x: x)
    else:
        raise ValueError(f"Unknown algorithm: {algorithm}")


@pvm(name="Random Consensus Builder")
def random_consensus_builder(profile, curr_cands=None, beta=0.5):
    """Random Consensus Builder (RCB) voting method due to Charikar et al. (https://arxiv.org/abs/2306.17838).

    For each ranking type in the profile, runs the deterministic Consensus Builder voting method using that ranking
    as the consensus building ranking. The probability of a candidate winning is proportional to the
    number of voters with rankings that would make that candidate win when used as the consensus
    building ranking.

    Args:
        profile (Profile): An anonymous profile of linear orders
        curr_cands (List[int], optional): Candidates to consider. Defaults to all candidates if not provided.
        beta (float): Threshold for elimination (default 0.5). When processing candidate i, eliminates a candidate j
                    above i in the consensus building ranking if the proportion of voters preferring i to j is >= beta

    Returns:
        dict: Maps each candidate to their probability of winning under the RCB method

    .. seealso::
        :meth:`pref_voting.iterative_methods.consensus_builder`
        :meth:`pref_voting.stochastic_methods.random_consensus_builder_st`
"""
    from pref_voting.iterative_methods import consensus_builder

    if curr_cands is None:
        curr_cands = profile.candidates

    # Count how many times each ranking type produces each winner
    winner_counts = {c: 0 for c in curr_cands}

    # Process each unique ranking type
    for ranking_type in profile.ranking_types:
        # Count number of voters with this ranking type
        num_rankings_with_type = len([r for r in profile.rankings if r == ranking_type])
        winner = consensus_builder(profile, curr_cands=curr_cands,consensus_building_ranking=ranking_type, beta=beta)[0]
        winner_counts[winner] += num_rankings_with_type
        total_count += num_rankings_with_type

    # Convert counts to probabilities
    return {c: count/profile.num_voters for c, count in winner_counts.items()}


def create_probabilistic_method(vm):
    """
    Create a probabilistic voting method from a voting method.
    """
    
    from pref_voting.voting_method import VotingMethod
    if type(vm) != VotingMethod:
        raise TypeError("vm must be a VotingMethod object")
    
    def _pvm(profile, curr_cands=None, **kwargs):
        return vm.prob(profile, curr_cands=curr_cands, **kwargs)
    
    return ProbVotingMethod(_pvm, name=f'{vm.name} with Even Chance Tiebreaking')

def mixture(pvm1, pvm2, alpha):
    """
    Mixture of the two probabilistic voting methods pvm1 and pvm2 with mixing parameter alpha. With probability alpha, the output is the output of pvm1, and with probability 1-alpha, the output is the output of pvm2.
    """
    def _mixture(profile, curr_cands=None, **kwargs):
        return {c: alpha * pvm1(profile, curr_cands=curr_cands, **kwargs)[c] + (1-alpha) * pvm2(profile, curr_cands=curr_cands, **kwargs)[c] for c in profile.candidates}
    
    return ProbVotingMethod(_mixture, name=f'Mixture of {pvm1.name} and {pvm2.name} with alpha={alpha}')


'''
    File: generate_margin_graphs.py
    Author: Wes Holliday (wesholliday@berkeley.edu) and Eric Pacuit (epacuit@umd.edu)
    Date: July 14, 2022
    Updated: December 19, 2022
    
    Functions to generate a margin graph
    
'''


import networkx as nx
from itertools import combinations
from pref_voting.helper import sublists, compositions, enumerate_compositions, convex_lexicographic_sublists
from pref_voting.weighted_majority_graphs import MarginGraph
import numpy as np
from scipy.stats import multivariate_normal

def generate_edge_ordered_tournament(num_cands, parity="even"): 
    """Generate a random uniquely weighted MarginGraph for ``num_cands`` candidates.  

    :param num_cands: the number of candidates
    :type num_cands: int
    :returns: a uniquely weighted margin graph
    :rtype: MarginGraph

    .. note:: This function randomly generates a tournament with a linear order over the edges.  A **tournament** is an asymmetric directed graph with an edge between every two nodes.  The linear order of the edges is represented by assigning to each edge a number  :math:`2, \ldots, 2*n`, where :math:`n` is the number of the edges. 
    """

    assert parity in ["even", "odd"], "The parity should be either 'even' or 'odd'."

    mg = nx.DiGraph()
    mg.add_nodes_from(range(num_cands))
    _edges = list()
    for c1 in range(num_cands): 
        for c2 in range(c1+1, num_cands): 
            if np.random.choice([True, False]): 
                _edges.append((c1, c2))
            else: 
                _edges.append((c2, c1))
                   
    edges = list()
    edge_indices = list(range(len(_edges)))
    np.random.shuffle(edge_indices)
    
    for i, e_idx in enumerate(edge_indices):
        edges.append((_edges[e_idx][0], _edges[e_idx][1], 2 * (i+1) if parity == 'even' else 2 * i+1)) 
    
    return MarginGraph(range(num_cands), edges)

def generate_margin_graph(num_cands, weight_domain = None, parity = 'even'): 
    """Generate a random MarginGraph (allowing for ties in the margins) for ``num_cands`` candidates.  

    Args:
        num_cands (int): the number of candidates

    Returns:
        MarginGraph

    """

    assert parity in ['even', 'odd'], "Parity must be 'even' or 'odd'."
    assert weight_domain is None or isinstance(weight_domain, list) and len(weight_domain) > 0, "The weight_domain must be a list with at least one element."
    
    candidates = list(range(num_cands))
    edges = list()
    pairs_of_cands = list(combinations(candidates, 2))

    if weight_domain is None and parity == 'even': 
        weight_domain = [2 * pidx for pidx in range(len(pairs_of_cands) + 1)]
    elif weight_domain is None and parity == 'odd': 
        weight_domain = [2 * pidx + 1 for pidx in range(len(pairs_of_cands) + 1)]
    # 

    for c1, c2 in pairs_of_cands:

        margin = np.random.choice(weight_domain)
        
        if margin != 0: 
            if np.random.choice([True, False]): 
                edges.append((c1, c2, margin))
            else:
                edges.append((c2, c1, margin))

    return MarginGraph(candidates, edges)

def generate_margin_graph_bradley_terry(num_cands, num_voters, score_prob_mod = lambda c: np.random.uniform(0,1)):
    """Generates a margin graph for num_cands candidates by first sampling candidate scores from score_prob_mod and then sampling votes from the Bradley-Terry model using the sampled scores.

    Args:
        num_cands (int): Number of candidates
        num_voters (int): Number of voters
        score_prob_mod (function, optional): A function that takes a candidate and returns a score. Defaults to lambda c: np.random.uniform(0,1).

    Returns:
        MarginGraph: A margin graph
    """

    candidates = list(range(num_cands))
    pairs_of_cands = list(combinations(candidates, 2))

    cand_score = dict()
    for c in candidates:
        cand_score[c] = score_prob_mod(c)

    edges = list()
    
    for c1, c2 in pairs_of_cands:

        support_c1_c2 = 0
        support_c2_c1 = 0

        for n in range(num_voters):
            vote = np.random.choice([1,0], p = [cand_score[c1] / (cand_score[c1] + cand_score[c2]), cand_score[c2] / (cand_score[c1] + cand_score[c2])])

            if vote == 1:
                support_c1_c2 += 1
            else:
                support_c2_c1 += 1
        
        if support_c1_c2 > support_c2_c1:
            edges.append((c1,c2, support_c1_c2 - support_c2_c1))

        if support_c2_c1 > support_c1_c2:
            edges.append((c2,c1, support_c2_c1 - support_c1_c2))

    return MarginGraph(candidates, edges)

### 

# Turn a code into a pair
def depair(pair_vector, k):
    return pair_vector[k]

# This function defines the i,jth entry of the covariance matrix
def entries(pair_vector, i,j):
    x = depair(pair_vector, i)
    y = depair(pair_vector, j)
    if x[0] == y[0] and x[1] == y[1]:
        return 1
    if x[1] == y[0]:
        return -1/3
    if x[1] == y[1]:
        return 1/3
    if x[0] == y[0]:
        return 1/3
    if x[0] == y[1]:
        return -1/3
    return 0

def generate_covariance_matrix(num_candidates):
    
    num_pairs = num_candidates *(num_candidates -1)//2

    # Store the vector mapping codes to pairs
    pair_vector = [0]*num_pairs

    # Populate the vector of pairs
    k=0
    for i in range(num_candidates):
        for j in range(i+1,num_candidates):
            pair_vector[k] = [i,j]
            k = k+1

    # Populate the covariance matrix
    cov = np.empty((num_pairs,num_pairs))
    for i in range(num_pairs):
        for j in range(num_pairs):
            cov[i,j] = entries(pair_vector, i,j)
            
    return cov


def generate_edge_ordered_tournament_infinite_limit(num_candidates, cov_matrix = None): 
    """
    Using the ideas from Section 9 of the paper 
    *An Analysis of Random Elections with Large Numbers of Voters* by Matthew Harrison-Trainor 
    (https://arxiv.org/abs/2009.02979) and the code provided at  
    https://github.com/MatthewHT/RandomMarginGraphs/, generate a qualitative margin graph for 
    ``num_candidates`` candidates.
    
    .. important:: 
        
        The weights of the generated margin graphs are real numbers, representing a linear ordering of the edges. 
        Only qualitative margin graph invariant voting methods, such as Split Cycle, Beat Path, Minimax, 
        Ranked Pairs, etc., should be used on the generated graphs. 
        
    Args:
        
        num_candidates (int): the number of candidates
        
    Returns: 
    
        MarginGraph
    
    """
    
    candidates = range(num_candidates)
    cov_matrix = cov_matrix if cov_matrix is not None else generate_covariance_matrix(num_candidates)
    # random_var is a random variable with the multivariate normal distribution of margin graphs
    random_var = multivariate_normal(None, cov_matrix)
    rv = random_var.rvs()
    
    def pair(p):
        return p[1]-2*p[0]-1 + (num_candidates)*(num_candidates+1)//2 - (num_candidates-p[0])*(num_candidates-p[0]+1)//2

    mg = [[-np.inf for _ in candidates] for _ in candidates]
    
    for c1 in candidates:
        for c2 in candidates:
            if (c1 < c2 and rv[pair([c1,c2])] > 0):
                mg[c1][c2] = rv[pair([c1,c2])]
            if (c1 > c2 and rv[pair([c2,c1])] < 0):
                mg[c1][c2] = -rv[pair([c2,c1])]
            if (c1 == c2):
                mg[c1][c2] = 0

    w_edges = [(c1, c2, mg[c1][c2]) 
               for c1 in candidates 
               for c2 in candidates if c1 != c2 if mg[c1][c2] > 0]
    
    return MarginGraph(candidates, w_edges)

## Generating Canonical MarginGraphs without Tied Margins

def _enumerate_ceots(num_cands, num_edges, partial_ceot, used_nodes, next_node):

    # Given a partial ceot P, we can extend it with any new edge (A,B) satisfying one of the following conditions:

    # 1. A and B have both already been used in edges in P, but neither (A,B) nor (B,A) is in P;
    # 2. A has already been used in an edge in P, and B is the next integer after the largest integer in an edge in P.
    # 3. A is the next integer after the largest integer in an edge in P, and B has already been used in an edge in P;
    # 4. A is the next integer after the largest integer in an edge in P, and B is the next integer after A.

    if len(partial_ceot) == num_edges:
        yield partial_ceot

    if len(partial_ceot) < num_edges:

        if next_node == num_cands:
            available_nodes = used_nodes

        if next_node < num_cands:
            available_nodes = used_nodes + [next_node]

        for n in available_nodes:

            if n == next_node and next_node < num_cands - 1: # If n == next_node, we are in Case 3 or Case 4 above
                
                available_nodes = used_nodes + [next_node + 1]

            for m in available_nodes:
                
                if not n==m and not (n,m) in partial_ceot and not (m,n) in partial_ceot:
                
                    new_ceot = [edge for edge in partial_ceot] + [(n,m)]

                    if not (n == next_node or m == next_node): # Then we are in Case 1 above

                        yield from _enumerate_ceots(num_cands,num_edges,new_ceot,used_nodes,next_node)

                    if (n == next_node or m == next_node) and not m == next_node + 1: # Then we are in Case 2 or 3 above

                        new_used_nodes = list(set(used_nodes + [n,m]))
                        new_next_node = next_node + 1
                        
                        yield from _enumerate_ceots(num_cands,num_edges,new_ceot,new_used_nodes,new_next_node)

                    if m == next_node + 1: # Then we are in Case 4 above

                        new_used_nodes = list(set(used_nodes + [n,m]))
                        new_next_node = next_node + 2
                        
                        yield from _enumerate_ceots(num_cands,num_edges,new_ceot,new_used_nodes,new_next_node)

def _enumerate_ceots_as_edgelist(num_cands):

    num_edges = (num_cands * (num_cands -1))//2

    partial_ceot = [(0,1)]

    used_nodes = [0,1]

    next_node = 2

    yield from _enumerate_ceots(num_cands,num_edges,partial_ceot,used_nodes,next_node)

    
def enumerate_canonical_edge_ordered_tournaments(num_cands, parity = "even"):  
    """
    A *canonical* edge-ordered tournament (ceot) is a representative from an isomorphism class of  
    linearly edge-ordered tournaments.  Enumerate all ceots for ``num_cands`` candidates, representing 
    a ceot as a ``MarginGraph`` where the margins represent the linear order of the edges.  
    
    Args:
        num_cands (int): the number of candidates
        parity (str, optional): The parity of the margins, either 'even' or 'odd'.

    Returns:
        A generator of ``MarginGraph`` for ``num_candidates``
        
    .. warning:: It is only feasible to finish the enumeration for up to 5 candidates.    

    """
    
    assert parity in ["odd", "even"], "parity must be either 'odd' or 'even'"
    
    for ceot in _enumerate_ceots_as_edgelist(num_cands): 
        yield MarginGraph(list(range(num_cands)), 
                          [(e[0], e[1], 2 * (eidx + 1) if parity == "even" else 2 * eidx + 1) 
                           for eidx, e in enumerate(reversed(ceot))])

            
def enumerate_uniquely_weighted_margin_graphs(num_cands, weight_domain):  
    """
    Enumerate all representatives from isomorphism classes of uniquely-weighted margin graphs with weights drawn from ``weight_domain``. 
    
    Args:
        num_cands (int): the number of candidates
        weight_domain (List[int]): The list of weights in the margin graph.
        
    Returns:
        A generator of ``MarginGraph`` for ``num_candidates``
        

    .. warning:: It is only feasible to finish the enumeration for up to 5 candidates.    

    """
    
    weight_domain = sorted(weight_domain)
    
    num_edges = (num_cands * (num_cands - 1)) // 2
    
    for ceot in _enumerate_ceots_as_edgelist(num_cands): 
        
        for weight_list in sublists(weight_domain, num_edges): 
            yield MarginGraph(list(range(num_cands)), 
                              [(e[0], e[1], weight_list[eidx]) for eidx, e in enumerate(reversed(ceot))])


## Generating Canonical MarginGraphs with Tied Margins

def _enumerate_cweots_as_edgelist(num_cands, include_weak_tournaments=True):
    
    #Enumerate each canonical weakly edge ordered tournament as a list of lists of tied edges.
    #If include_weak_tournaments = True, then allow weak tournaments in which two nodes may have no edge between them.   
    
    def edge_match(e1, e2):
        return e1['weight'] == e2['weight']

    cweots = dict() # For isomorphism checking, keep track of the cweots generated so far.
    
    if include_weak_tournaments:
        cweots_with_absent_edges = dict() # For isomorphism checking, keep track of the cweots for which some edges are absent.

    for ceot in tqdm(list(_enumerate_ceots_as_edgelist(num_cands))):

        # The sorted list of number of wins by each candidate will be a useful invariant for isomorphism checking below.
        win_vector = tuple(sorted([len([edge for edge in ceot if edge[0] == i]) for i in range(num_cands)]))
        
        # Given ceot, we will generate many cweots as follows:

        # 1. Collect all the convex lexicographic sublists of ceot in order as [L1,...,Ln]. 
        # It suffices to only consider convex lexicographic sublists because for any cweot, 
        # we can obtain a ceot by breaking all ties between edges in a tied group lexicographically.

        l_sublists = convex_lexicographic_sublists(ceot)

        # 2. Within each L_i, we want to consider all ways of making consecutive edges tied.
        # Such a way is given by a composition of the integer len(L_i). 
        # Thus, we first iterate over all compositions of len(L_i),...,len(L_n).

        int_list = [len(s) for s in l_sublists]

        # 3. Since the above approach overgenerates cweots, we will check for isomorphism before adding a cweot to our list.

        for compositions in enumerate_compositions(int_list):

            cases = [False, True] if include_weak_tournaments else [False]

            for consider_weak_tourns in cases:
                
                cweot = []

                for idx, s in enumerate(l_sublists):
                    
                    composition = compositions[idx]
    
                    for n in composition:
                        cweot.append(s[:n])
                        s=s[n:]

                # If we are considering weak tournaments in this case, we remove the last tied group of edges and compute the sorted win-loss vector.
                if consider_weak_tourns:
                    win_loss_vector = tuple(sorted([(len([edge for edge in ceot if edge[0] == i and edge not in cweot[-1]]),len([edge for edge in ceot if edge[1] == i and edge not in cweot[-1]])) for i in range(num_cands)]))
                    cweot = cweot[:-1]
                    
                G = nx.DiGraph()
                weight = len(ceot)
                for group in cweot:
                    for edge in group:
                        G.add_edge(edge[0], edge[1], weight=weight)
                    weight = weight-1

                add_graph = True
                
                # Next we check whether G is isomorphic to a cweot G2 already generated.
                # We only need to check those cweots G2 that have (i) the same sorted Copeland scores and 
                # (ii) the same list of numbers of edges in each tied group as G, 
                # since these are necessary conditions for isomorphism.

                tie_sizes = tuple([len(s) for s in cweot])

                if not consider_weak_tourns:

                    invariant = (win_vector, tie_sizes)
                    
                    if invariant not in cweots.keys():
                        cweots[invariant] = []
              
                    for idx, G2 in enumerate(cweots[invariant]):
                        if nx.is_isomorphic(G, G2, edge_match=edge_match):  
                            add_graph = False
                            break 

                    if add_graph:   
                        cweots[invariant].append(G)    
                        yield cweot

                if consider_weak_tourns:

                    invariant = (win_loss_vector, tie_sizes)

                    if invariant not in cweots_with_absent_edges.keys():
                        cweots_with_absent_edges[invariant] = []

                    for idx, G2 in enumerate(cweots_with_absent_edges[invariant]):
                        if nx.is_isomorphic(G, G2, edge_match=edge_match):  
                            add_graph = False
                            break

                    if add_graph:
                        cweots_with_absent_edges[invariant].append(G)
                        yield cweot

def enumerate_canonical_weakly_edge_ordered_tournaments(num_cands, parity = "even", include_weak_tournaments = True):  
    """
    A *canonical* weakly edge-ordered tournament (cweot) is a representative from an isomorphism class of  
    weakly edge-ordered tournaments.  Enumerate all cweots for ``num_cands`` candidates, representing 
    a cweot as a ``MarginGraph`` where the margins represent the order of the edges.

    If include_weak_tournaments = True, then allow weak tournaments in which two nodes may have no edge between them.
    
    Args:
        num_cands (int): the number of candidates
        parity (str, optional): The parity of the margins, either 'even' or 'odd'.

    Returns:
        A generator of ``MarginGraph`` for ``num_candidates``
        
    .. warning:: It is only feasible to finish the enumeration for up to 4 candidates.    

    """
    
    assert parity in ["odd", "even"], "parity must be either 'odd' or 'even'"
    
    for cweot in _enumerate_cweots_as_edgelist(num_cands, include_weak_tournaments=include_weak_tournaments): 

        weighted_edges = list()

        for idx, group in enumerate(reversed(cweot)):
            for e in group:
                weighted_edge = (e[0], e[1], 2 * (idx + 1) if parity == "even" else 2 * idx + 1) 
                weighted_edges.append(weighted_edge)

        yield MarginGraph(list(range(num_cands)), weighted_edges)

def enumerate_margin_graphs(num_cands, weight_domain, include_weak_tournaments = True):  
    """
    Enumerate all representatives from isomorphism classes of margin graphs with weights drawn from ``weight_domain``. 
    
    Args:
        num_cands (int): the number of candidates
        weight_domain (List[int]): The list of weights in the margin graph.
        
    Returns:
        A generator of ``MarginGraph`` for ``num_candidates``
        
    .. warning:: It is only feasible to finish the enumeration for up to 4 candidates.    

    """
    
    weight_domain = sorted(weight_domain)
    
    for cweot in _enumerate_cweots_as_edgelist(num_cands, include_weak_tournaments = include_weak_tournaments): 

        for weight_list in sublists(weight_domain, len(cweot)):

            weighted_edges = list()

            for idx, group in enumerate(reversed(cweot)):
                for e in group:
                    weighted_edge = (e[0], e[1], weight_list[idx]) 
                    weighted_edges.append(weighted_edge)
            
            yield MarginGraph(list(range(num_cands)), weighted_edges)


"""
    File: axiom.py
    Author: Wes Holliday (wesholliday@berkeley.edu) and Eric Pacuit (epacuit@umd.edu)
    Date: April 27, 2023
    
    Define the Axiom class. 
"""

class Axiom(object): 
    """
    A class to represent voting method axioms. 

    Args:
        name (string): The human-readable name of the axiom.
        has_violation (function): function that returns a Boolean which is True when there is a violation of the axiom.
        find_all_violations (function): function that returns all instances of violations of the axiom.
        satisfying_vms (list): list of voting methods satisfying the axiom.
        violating_vms (list): list of voting methods violating the axiom.

    """
    def __init__(self, name, has_violation, find_all_violations):
        self.name = name
        self.has_violation = has_violation
        self.find_all_violations = find_all_violations
        self.satisfying_vms = list()
        self.violating_vms = list()
        
    def satisfies(vm): 
        return vm.name in self.satisfying_vms
        
    def violates(vm): 
        return vm.name in self.violating_vms
        
    def add_satisfying_vms(vms): 
        self.satisfying_vms += vms

    def add_violating_vms(vms): 
        self.violating_vms += vms




from pref_voting.profiles import Profile
from pref_voting.profiles_with_ties import ProfileWithTies
from pref_voting.rankings import Ranking
from itertools import combinations, chain, permutations
import copy

def display_mg(edata): 
    if type(edata) == Profile or type(edata) == ProfileWithTies: 
        edata.display_margin_graph()
    else: 
        edata.display()

def list_to_string(cands, cmap): 
    return "{" + ', '.join([cmap[c] for c in cands]) + "}"


def swap_candidates(ranking, c1, c2):
    """
    Swap two candidates in a ranking.
    :param ranking: either a tuple or a list of candidates or a Ranking object
    :param c1: candidate 1
    :param c2: candidate 2
    :return: a new ranking (a tuple) with c1 and c2 swapped
    """

    if isinstance(ranking, Ranking):

        rmap = ranking.rmap

        if c1 not in rmap or c2 not in rmap:
            raise ValueError("One of the candidates is not in the ranking")
        
        # swap the values associated with c1 and c2
        new_rmap = rmap.copy()
        new_rmap[c1], new_rmap[c2] = new_rmap[c2], new_rmap[c1]
        new_ranking = Ranking(new_rmap)
        
    elif isinstance(ranking, (list, tuple)):
        new_ranking = []
        for c in ranking:
            if c == c1:
                new_ranking.append(c2)
            elif c == c2:
                new_ranking.append(c1)
            else:
                new_ranking.append(c)
        new_ranking = tuple(new_ranking)
    return new_ranking



def equal_size_partitions_with_duplicates(lst):
    """
    Generate all partitions of a list into two distinct subsets of equal size, 
    including cases where the input list contains duplicates and elements 
    that do not support ordering (<).
    
    Parameters:
        lst (list): The input list to partition. Must have an even number of elements.
        
    Returns:
        list of tuples: A list of tuples, where each tuple contains two lists of equal size.
    """
    if len(lst) % 2 != 0:
        raise ValueError("The input list must have an even number of elements.")
    
    n = len(lst) // 2
    partitions = []
    
    seen = set()
    
    for subset in combinations(lst, n):
        complement = lst[:]
        for item in subset:
            complement.remove(item)
        
        partition_key = frozenset([frozenset(subset), frozenset(complement)])
        if partition_key not in seen:
            seen.add(partition_key)
            partitions.append((list(subset), complement))
    
    return partitions


def get_rank(ranking, c):
    """
    Get the (normalized) rank of a candidate in a ranking.
    :param ranking: either a tuple or a list of candidates or a Ranking object
    :param c: candidate
    :return: the rank of c in the ranking
    """
    if isinstance(ranking, Ranking):
        norm_ranking = copy.deepcopy(ranking)
        norm_ranking.normalize_ranks()
        return norm_ranking.rmap[c]
    elif isinstance(ranking, (list, tuple)):
        return ranking.index(c)
    else:
        raise ValueError("Invalid input type")
    

# generate all subsets of a set, use combinations
def powerset(lst):
    s = list(lst)
    return chain.from_iterable(combinations(s, r) for r in range(len(s)+1))

def linear_orders_with_reverse(cands): 

    lin_orders = list(permutations(cands))
    lin_orders_with_reverse = []
    for lin_order in lin_orders:
        lin_orders_with_reverse.append((lin_order, lin_order[::-1]))
    return lin_orders_with_reverse

def remove_first_occurrences(rankings, r1, r2):
    removed_r1 = False
    removed_r2 = False
    result = []

    for r in rankings:
        if r == r1 and not removed_r1:
            removed_r1 = True  # Skip the first r1
        elif r == r2 and not removed_r2:
            removed_r2 = True  # Skip the first r2
        else:
            result.append(r)  # Keep all other elements

    return result



"""
    File: invariance_axioms.py
    Author: Wesley H. Holliday (wesholliday@berkeley.edu) and Eric Pacuit (epacuit@umd.edu)
    Date: January 11, 2024
    Updated: May 29, 2025
    
    Invariance axioms 
"""

from pref_voting.axiom import Axiom
from pref_voting.axiom_helpers import *
from itertools import permutations
from pref_voting.profiles import Profile


def _homogeneity_violation(edata, vm, num_copies, violation_type, verbose=False):

    ws = vm(edata)

    if isinstance(edata, Profile):
        rankings, old_rcounts = edata.rankings_counts

        new_rcounts = [c * num_copies for c in old_rcounts]

        new_edata = Profile(list(map(list, rankings)), rcounts = new_rcounts, cmap = edata.cmap)
    else:
        old_rankings, old_rcounts = edata.rankings_counts

        new_rcounts = [c * num_copies for c in old_rcounts]

        new_edata = ProfileWithTies(old_rankings, rcounts = new_rcounts, candidates = edata.candidates, cmap = edata.cmap)

        if edata.using_extended_strict_preference:
            new_edata.use_extended_strict_preference()
    
    new_ws = vm(new_edata)

    violation = False

    if violation_type == "Homogeneity":
        if new_ws != ws:
            violation = True
            return_value = list(set(new_ws) ^ set(ws))
    
    if violation_type == "Upward Homogeneity":
        if any([w for w in ws if w not in new_ws]):
            violation = True
            return_value = [w for w in ws if w not in new_ws]

    if violation_type == "Downward Homogeneity":
        if any([w for w in new_ws if w not in ws]):
            violation

    if violation:

        if verbose:
            print(f"{violation_type} Violation for {vm.name}")
            edata.display()
            print(edata.description())
            vm.display(edata)
            print("")

            new_edata.display()
            print(new_edata.description())
            vm.display(new_edata)
 
        return True, return_value
    
    return False, list()

def has_homogeneity_violation(edata, vm, num_copies = 2, verbose=False):
    """
    Returns True if replacing each ranking with num_copies of that ranking changes the set of winners.
    
    Args:
        edata (Profile, ProfileWithTies): the election data.
        vm (VotingMethod): A voting method to test.
        num_copies (int, default=2): The number of copies to multiply each ranking by.
        verbose (bool, default=False): If a violation is found, display the violation. 

    Returns: 
        Result of the test (bool): Returns True if there is a violation and False otherwise. 

    """
    return _homogeneity_violation(edata, vm, num_copies, "Homogeneity", verbose=verbose)[0]

def find_all_homogeneity_violations(edata, vm, num_copies = 2, verbose=False):
    """
    Returns the symmetric difference of the winners before and after multiplying the number of copies of each ranking.
    
    Args:
        edata (Profile, ProfileWithTies): the election data.
        vm (VotingMethod): A voting method to test.
        num_copies (int, default=2): The number of copies to multiply each ranking by.
        verbose (bool, default=False): If a violation is found, display the violation. 

    Returns: 
        return_value: A list of the symmetric difference between the old and new winners.

    """
    return _homogeneity_violation(edata, vm, num_copies, "Homogeneity", verbose=verbose)[1]

homogeneity = Axiom(
    "Homogeneity",
    has_violation = has_homogeneity_violation,
    find_all_violations = find_all_homogeneity_violations, 
)

def has_upward_homogeneity_violation(edata, vm, num_copies = 2, verbose=False):
    """
    Returns True if replacing each ranking with num_copies of that ranking causes some winner to lose.
    
    Args:
        edata (Profile, ProfileWithTies): the election data.
        vm (VotingMethod): A voting method to test.
        num_copies (int, default=2): The number of copies to multiply each ranking by.
        verbose (bool, default=False): If a violation is found, display the violation. 

    Returns: 
        Result of the test (bool): Returns True if there is a violation and False otherwise. 

    """
    return _homogeneity_violation(edata, vm, num_copies, "Upward Homogeneity", verbose=verbose)[0]

def find_all_upward_homogeneity_violations(edata, vm, num_copies = 2, verbose=False):
    """
    Returns the set of winners who lose as a result of replacing each ranking with num_copies of that ranking.
    
    Args:
        edata (Profile, ProfileWithTies): the election data.
        vm (VotingMethod): A voting method to test.
        num_copies (int, default=2): The number of copies to multiply each ranking by.
        verbose (bool, default=False): If a violation is found, display the violation. 

    Returns: 
        return_value: the set of winners who lose as a result of replacing each ranking with num_copies of that ranking.

    """
    return _homogeneity_violation(edata, vm, num_copies, "Upward Homogeneity", verbose=verbose)[1]

upward_homogeneity = Axiom(
    "Upward Homogeneity",
    has_violation = has_upward_homogeneity_violation,
    find_all_violations = find_all_upward_homogeneity_violations, 
)

def has_downward_homogeneity_violation(edata, vm, num_copies = 2, verbose=False):
    """
    Returns True if replacing each ranking with num_copies of that ranking causes some loser to win.
    
    Args:
        edata (Profile, ProfileWithTies): the election data.
        vm (VotingMethod): A voting method to test.
        num_copies (int, default=2): The number of copies to multiply each ranking by.
        verbose (bool, default=False): If a violation is found, display the violation. 

    Returns: 
        Result of the test (bool): Returns True if there is a violation and False otherwise. 

    """
    return _homogeneity_violation(edata, vm, num_copies, "Downward Homogeneity", verbose=verbose)[0]

def find_all_downward_homogeneity_violations(edata, vm, num_copies = 2, verbose=False):
    """
    Returns the set of losers who win as a result of replacing each ranking with num_copies of that ranking.
    
    Args:
        edata (Profile, ProfileWithTies): the election data.
        vm (VotingMethod): A voting method to test.
        num_copies (int, default=2): The number of copies to multiply each ranking by.
        verbose (bool, default=False): If a violation is found, display the violation. 

    Returns: 
        return_value: the set of losers who win as a result of replacing each ranking with num_copies of that ranking.

    """
    return _homogeneity_violation(edata, vm, num_copies, "Downward Homogeneity", verbose=verbose)[1]

downward_homogeneity = Axiom(
    "Downward Homogeneity",
    has_violation = has_downward_homogeneity_violation,
    find_all_violations = find_all_downward_homogeneity_violations, 
)


def _has_block_violation(edata, vm, violation_type, verbose=False):

    ws = vm(edata)

    if isinstance(edata, Profile):
        old_rankings, old_rcounts = edata.rankings_counts

        new_rankings = list(permutations(edata.candidates))
        new_rcounts = [1] * len(new_rankings)

        new_edata = Profile(list(map(list, old_rankings)) + new_rankings, rcounts = list(old_rcounts) + new_rcounts, cmap = edata.cmap)
    else:
        old_rankings, old_rcounts = edata.rankings_counts

        new_rankings = [{c: perm.index(c)+1 for c in edata.candidates} for perm in permutations(edata.candidates)]
        new_rcounts = [1] * len(new_rankings)

        new_edata = ProfileWithTies(old_rankings + new_rankings, rcounts = old_rcounts + new_rcounts, candidates = edata.candidates, cmap = edata.cmap)
    
    new_ws = vm(new_edata)

    violation = False

    if violation_type == "Block Invariance":
        if new_ws != ws:
            violation = True
            return_value = list(set(new_ws) ^ set(ws))
    
    if violation_type == "Upward Block Preservation":
        if any([w for w in ws if w not in new_ws]):
            violation = True
            return_value = [w for w in ws if w not in new_ws]

    if violation_type == "Downward Block Preservation":
        if any([w for w in new_ws if w not in ws]):
            violation = True
            return_value = [w for w in new_ws if w not in ws]

    if violation:

        if verbose:
            print(f"{violation_type} Violation for {vm.name}")
            edata.display()
            print(edata.description())
            vm.display(edata)

            new_edata.display()
            print(new_edata.description())
            vm.display(new_edata)
 
        return True, return_value
    
    return False, list()

def has_block_invariance_violation(edata, vm, verbose=False):
    """
    Returns True if adding a block of all linear orders changes the set of winners.
    
    Args:
        edata (Profile, ProfileWithTies): the election data.
        vm (VotingMethod): A voting method to test.
        verbose (bool, default=False): If a violation is found, display the violation. 

    Returns: 
        Result of the test (bool): Returns True if there is a violation and False otherwise. 

    """
    return _has_block_violation(edata, vm, "Block Invariance", verbose=verbose)[0]

def find_all_block_invariance_violations(edata, vm, verbose=False):
    """
    Returns the symmetric difference of the winners before and after adding a block of all linear orders.
    
    Args:
        edata (Profile, ProfileWithTies): the election data.
        vm (VotingMethod): A voting method to test.
        verbose (bool, default=False): If a violation is found, display the violation. 

    Returns: 
        return_value: A list of the symmetric difference between the old and new winners.

    """
    return _has_block_violation(edata, vm, "Block Invariance", verbose=verbose)[1]

block_invariance = Axiom(
    "Block Invariance",
    has_violation = has_block_invariance_violation,
    find_all_violations = find_all_block_invariance_violations, 
)

def has_upward_block_preservation_violation(edata, vm, verbose=False):
    """
    Returns True if adding a block of all linear orders causes some winner to lose.
    
    Args:
        edata (Profile, ProfileWithTies): the election data.
        vm (VotingMethod): A voting method to test.
        verbose (bool, default=False): If a violation is found, display the violation. 

    Returns: 
        Result of the test (bool): Returns True if there is a violation and False otherwise. 

    """
    return _has_block_violation(edata, vm, "Upward Block Preservation", verbose=verbose)[0]

def find_all_upward_block_preservation_violations(edata, vm, verbose=False):
    """
    Returns the set of winners who lose as a result of adding a block of all linear orders.
    
    Args:
        edata (Profile, ProfileWithTies): the election data.
        vm (VotingMethod): A voting method to test.
        verbose (bool, default=False): If a violation is found, display the violation. 

    Returns: 
        return_value: the set of winners who lose as a result of adding a block of all linear orders.

    """
    return _has_block_violation(edata, vm, "Upward Block Preservation", verbose=verbose)[1]

upward_block_preservation = Axiom(
    "Upward Block Preservation",
    has_violation = has_upward_block_preservation_violation,
    find_all_violations = find_all_upward_block_preservation_violations, 
)

def has_downward_block_preservation_violation(edata, vm, verbose=False):
    """
    Returns True if adding a block of all linear orders causes some loser to win.
    
    Args:
        edata (Profile, ProfileWithTies): the election data.
        vm (VotingMethod): A voting method to test.
        verbose (bool, default=False): If a violation is found, display the violation. 

    Returns: 
        Result of the test (bool): Returns True if there is a violation and False otherwise. 

    """
    return _has_block_violation(edata, vm, "Downward Block Preservation", verbose=verbose)[0]

def find_all_downward_block_preservation_violations(edata, vm, verbose=False):
    """
    Returns the set of losers who win as a result of adding a block of all linear orders.
    
    Args:
        edata (Profile, ProfileWithTies): the election data.
        vm (VotingMethod): A voting method to test.
        verbose (bool, default=False): If a violation is found, display the violation. 

    Returns: 
        return_value: the set of losers who win as a result of adding a block of all linear orders.

    """
    return _has_block_violation(edata, vm, "Downward Block Preservation", verbose=verbose)[1]

downward_block_preservation = Axiom(
    "Downward Block Preservation",
    has_violation = has_downward_block_preservation_violation,
    find_all_violations = find_all_downward_block_preservation_violations, 
)

def has_preferential_equality_violation(prof, vm, verbose=False):
    """
    Check if a profile has a preferential equality violation for the voting method vm.

    See Definition 2.1 and Lemma 2.4 from the paper "Characterizations of voting rules based on majority margins" by Y. Ding,  W. Holliday, and E. Pacuit.

    """
    if isinstance(prof, ProfileWithTies):
       prof_constructor = ProfileWithTies
       prof = prof.add_unranked_candidates()
    
    else:
        prof_constructor = Profile
    
    for x, y in combinations(prof.candidates, 2):
        
        xy_rankings = [r for r in prof.rankings if get_rank(r, x) == get_rank(r, y) - 1 and not any(get_rank(r,z) == get_rank(r,x) for z in prof.candidates if z != x) and not any(get_rank(r,z) == get_rank(r,y) for z in prof.candidates if z != y)]

        other_rankings = [r for r in prof.rankings if r not in xy_rankings]

        if len(xy_rankings) != 0 and len(xy_rankings) % 2 == 0:
            for I, J in equal_size_partitions_with_duplicates(xy_rankings):
                new_rankings_I = [swap_candidates(r, x, y) for r in I] + list(J) + list(other_rankings)
                prof_I = prof_constructor(new_rankings_I)
                new_rankings_J = [swap_candidates(r, x, y) for r in J] + list(I) + list(other_rankings)
                prof_J = prof_constructor(new_rankings_J)

                if vm(prof_I) != vm(prof_J): 
                    if verbose:
                        print("The original profile")
                        prof.anonymize().display()
                        prof.display_margin_graph()
                        print(prof.description())
                        vm.display(prof)
                        if prof_constructor == ProfileWithTies:
                            print(f"\nThe profile after swapping {x} and {y} in the rankings {[r.rmap for r in I]}:")
                        else:
                            print(f"\nThe profile after swapping {x} and {y} in the rankings {[r for r in I]}:")
                        prof_I.anonymize().display()
                        prof_I.display_margin_graph()
                        print(prof_I.description())   
                        vm.display(prof_I)
                        if prof_constructor == ProfileWithTies:
                            print(f"\nThe profile after swapping {x} and {y} in the rankings {[r.rmap for r in J]}:")
                        else:
                            print(f"\nThe profile after swapping {x} and {y} in the rankings {[r for r in J]}:")
                        prof_J.anonymize().display()
                        prof_J.display_margin_graph()
                        print(prof_J.description())
                        vm.display(prof_J)
                    return True

        yx_rankings = [r for r in prof.rankings if get_rank(r, y) == get_rank(r, x) - 1 and not any(get_rank(r,z) == get_rank(r,x) for z in prof.candidates if z != x) and not any(get_rank(r,z) == get_rank(r,y) for z in prof.candidates if z != y)]

        other_rankings = [r for r in prof.rankings if r not in yx_rankings]

        if len(yx_rankings) != 0 and len(yx_rankings) % 2 == 0:
            for I, J in equal_size_partitions_with_duplicates(yx_rankings):
                new_rankings_I = [swap_candidates(r, y, x) for r in I] + list(J) + list(other_rankings)
                prof_I = prof_constructor(new_rankings_I)
                new_rankings_J = [swap_candidates(r, y, x) for r in J] + list(I) + list(other_rankings)
                prof_J = prof_constructor(new_rankings_J)

                if vm(prof_I) != vm(prof_J): 
                    if verbose:
                        print("The original profile")
                        prof.anonymize().display()
                        prof.display_margin_graph()
                        print(prof.description())
                        vm.display(prof)
                        if prof_constructor == ProfileWithTies:
                            print(f"\nThe profile after swapping {y} and {x} in the rankings {[r.rmap for r in I]}:")
                        else:
                            print(f"\nThe profile after swapping {y} and {x} in the rankings {[r for r in I]}:")
                        prof_I.anonymize().display()
                        prof_I.display_margin_graph()
                        print(prof_I.description())
                        vm.display(prof_I)
                        if prof_constructor == ProfileWithTies:
                            print(f"\nThe profile after swapping {y} and {x} in the rankings {[r.rmap for r in J]}:")
                        else:
                            print(f"\nThe profile after swapping {y} and {x} in the rankings {[r for r in J]}:")
                        prof_J.anonymize().display()
                        prof_J.display_margin_graph()
                        print(prof_J.description())
                        vm.display(prof_J)
                    return True
    return False

def find_all_preferential_equality_violations(prof, vm, verbose=False):
    """
    Return all the preferential equality violations for the voting method vm.  Returns a list of tuples of three profiles (prof, prof_I, prof_J) such that vm(prof_I) != vm(prof_J) and prof_I and prof_J are as defined Lemma 2.4 from the paper "Characterizations of voting rules based on majority margins" by Y. Ding,  W. Holliday, and E. Pacuit (see also Definition 2.1).

    """

    if isinstance(prof, ProfileWithTies):
       prof_constructor = ProfileWithTies
       prof = prof.add_unranked_candidates()
    
    else:
        prof_constructor = Profile
    
    violations = []
    for x, y in combinations(prof.candidates, 2):
        
        xy_rankings = [r for r in prof.rankings if get_rank(r, x) == get_rank(r, y) - 1 and not any(get_rank(r,z) == get_rank(r,x) for z in prof.candidates if z != x) and not any(get_rank(r,z) == get_rank(r,y) for z in prof.candidates if z != y)]

        other_rankings = [r for r in prof.rankings if r not in xy_rankings]

        if len(xy_rankings) != 0 and len(xy_rankings) % 2 == 0:
            for I, J in equal_size_partitions_with_duplicates(xy_rankings):
                new_rankings_I = [swap_candidates(r, x, y) for r in I] + list(J) + list(other_rankings)
                prof_I = prof_constructor(new_rankings_I)
                new_rankings_J = [swap_candidates(r, x, y) for r in J] + list(I) + list(other_rankings)   
                prof_J = prof_constructor(new_rankings_J)

                if vm(prof_I) != vm(prof_J): 
                    if verbose:
                        print("The original profile")
                        prof.anonymize().display()
                        prof.display_margin_graph()
                        print(prof.description())
                        vm.display(prof)
                        if prof_constructor == ProfileWithTies:
                            print(f"\nThe profile after swapping {x} and {y} in the rankings {[r.rmap for r in I]}:")
                        else:
                            print(f"\nThe profile after swapping {x} and {y} in the rankings {[r for r in I]}:")
                        prof_I.anonymize().display()
                        prof_I.display_margin_graph()
                        print(prof_I.description())
                        vm.display(prof_I)
                        if prof_constructor == ProfileWithTies:
                            print(f"\nThe profile after swapping {x} and {y} in the rankings {[r.rmap for r in J]}:")
                        else:
                            print(f"\nThe profile after swapping {x} and {y} in the rankings {[r for r in J]}:")
                        prof_J.anonymize().display()
                        prof_J.display_margin_graph()
                        print(prof_J.description())
                        vm.display(prof_J)
                    violations.append((prof, prof_I, prof_J))

        yx_rankings = [r for r in prof.rankings if get_rank(r, y) == get_rank(r, x) - 1 and not any(get_rank(r,z) == get_rank(r,x) for z in prof.candidates if z != x) and not any(get_rank(r,z) == get_rank(r,y) for z in prof.candidates if z != y)]

        other_rankings = [r for r in prof.rankings if r not in yx_rankings]

        if len(yx_rankings) != 0 and len(yx_rankings) % 2 == 0:
            for I, J in equal_size_partitions_with_duplicates(yx_rankings):
                new_rankings_I = [swap_candidates(r, y, x) for r in I] + list(J) + list(other_rankings)
                prof_I = prof_constructor(new_rankings_I)
                new_rankings_J = [swap_candidates(r, y, x) for r in J] + list(I) + list(other_rankings)
                prof_J = prof_constructor(new_rankings_J)

                if vm(prof_I) != vm(prof_J): 
                    if verbose:
                        print("The original profile")
                        prof.anonymize().display()
                        prof.display_margin_graph()
                        print(prof.description())
                        vm.display(prof)
                        if prof_constructor == ProfileWithTies:
                            print(f"\nThe profile after swapping {x} and {y} in the rankings {[r.rmap for r in I]}:")
                        else:
                            print(f"\nThe profile after swapping {x} and {y} in the rankings {[r for r in I]}:")
                        prof_I.anonymize().display()
                        prof_I.display_margin_graph()
                        print(prof_I.description())
                        vm.display(prof_I)
                        if prof_constructor == ProfileWithTies:
                            print(f"\nThe profile after swapping {x} and {y} in the rankings {[r.rmap for r in J]}:")
                        else:
                            print(f"\nThe profile after swapping {x} and {y} in the rankings {[r for r in J]}:")
                        prof_J.anonymize().display()
                        prof_J.display_margin_graph()
                        print(prof_J.description())
                        vm.display(prof_J)
                    violations.append((prof, prof_I, prof_J))
    return violations

preferential_equality = Axiom(
    "Preferential Equality",
    has_violation = has_preferential_equality_violation,
    find_all_violations = find_all_preferential_equality_violations, 
)


def has_tiebreaking_compensation_violation(prof, vm, verbose=False):
    """
    Return True if the profile prof has a tiebreaking compensation violation for the voting method vm.
    """
    if isinstance(prof, Profile):
        return False
    
    for cands in powerset(prof.candidates): 
        if len(cands) > 1: 

            rankings_with_tie = [r for r in prof.rankings if r.is_tied(cands)]

            checked_rankings = []
            for r1, r2 in combinations(rankings_with_tie, 2):
                if set([r1, r2]) in checked_rankings:
                    continue
                checked_rankings.append(set([r1, r2]))
                for lin_order, reverse_lin_order in linear_orders_with_reverse(cands): 

                    other_rankings = [r for r in prof.rankings if not r in rankings_with_tie]
                    
                    # rankings_with_tie without r1 and r2
                    other_rankings_with_tie = remove_first_occurrences(rankings_with_tie, r1, r2)

                    new_rankings = [r1.break_tie(lin_order),r2.break_tie(reverse_lin_order)] +  other_rankings_with_tie + other_rankings 

                    new_prof = ProfileWithTies(new_rankings, candidates=prof.candidates)
                    if vm(prof) != vm(new_prof): 
                        if verbose: 
                            prof.anonymize().display()
                            print(prof.description())
                            vm.display(prof)

                            print(f"\nAfter breaking the tie between the candidates {[prof.cmap[c] for c in cands]} in {r1} with {tuple([prof.cmap[c] for c in lin_order])} and {r2} with {tuple([prof.cmap[c] for c in reverse_lin_order])}: \n")
                            new_prof.anonymize().display()
                            print(new_prof.description())
                            vm.display(new_prof)
                        return True
    return False


def find_all_tiebreaking_compensation_violations(prof, vm, verbose=False):
    """
    Find all the violations of tiebreaking compensation for prof with respect to the voting method vm. Returns a list of tuples consisting of the rankings and the rankings with the ties broken. If there are no violations, return an empty list.
    """
    if isinstance(prof, Profile):
        return []

    violations = []
    for cands in powerset(prof.candidates): 
        if len(cands) > 1: 

            rankings_with_tie = [r for r in prof.rankings if r.is_tied(cands)]

            checked_rankings = []
            for r1, r2 in combinations(rankings_with_tie, 2):
                if set([r1, r2]) in checked_rankings:
                    continue
                checked_rankings.append(set([r1, r2]))
                for lin_order, reverse_lin_order in linear_orders_with_reverse(cands): 

                    other_rankings = [r for r in prof.rankings if not r in rankings_with_tie]
                    
                    # rankings_with_tie without r1 and r2
                    other_rankings_with_tie = remove_first_occurrences(rankings_with_tie, r1, r2)

                    new_rankings = [r1.break_tie(lin_order),r2.break_tie(reverse_lin_order)] +  other_rankings_with_tie + other_rankings 

                    new_prof = ProfileWithTies(new_rankings, candidates=prof.candidates)
                    if vm(prof) != vm(new_prof): 
                        if verbose: 
                            prof.anonymize().display()
                            print(prof.description())
                            vm.display(prof)

                            print(f"\nAfter breaking the tie between the candidates {[prof.cmap[c] for c in cands]} in {r1} with {tuple([prof.cmap[c] for c in lin_order])} and {r2} with {tuple([prof.cmap[c] for c in reverse_lin_order])}: \n")
                            new_prof.anonymize().display()
                            print(new_prof.description())
                            vm.display(new_prof)
                        violations.append(((r1, r1.break_tie(lin_order)), (r2, r2.break_tie(reverse_lin_order))))
    return violations

tiebreaking_compensation = Axiom(
    "Tiebreaking Compensation",
    has_violation = has_tiebreaking_compensation_violation,
    find_all_violations = find_all_tiebreaking_compensation_violations, 
)
invariance_axioms = [
    block_invariance,
    upward_block_preservation,
    downward_block_preservation,
    homogeneity,
    upward_homogeneity,
    downward_homogeneity,
    preferential_equality,
    tiebreaking_compensation,
]


import numpy as np
import matplotlib.pyplot as plt
import matplotlib.patheffects as path_effects
import seaborn as sns
from pref_voting.utility_functions import *
from pref_voting.utility_profiles import UtilityProfile

class SpatialProfile(object): 
    """
    A spatial profile is a set of candidates and voters in a multi-dimensional space.  Each voter and candidate is assigned vector of floats representing their position on each issue.

    Args:
        cand_pos (dict): A dictionary mapping each candidate to their position in the space.
        voter_pos (dict): A dictionary mapping each voter to their position in the space.

    Attributes:
        candidates (list): A list of candidates.
        voters (list): A list of voters.    
        cand_pos (dict): A dictionary mapping each candidate to their position in the space.    
        voter_pos (dict): A dictionary mapping each voter to their position in the space.   
        num_dims (int): The number of dimensions in the space.  
        cand_types (dict): A dictionary mapping each candidate to their type (e.g., party affiliation).

    """
    def __init__(self, cand_pos, voter_pos, candidate_types=None):

        cand_dims = [len(v) for v in cand_pos.values()]
        voter_dims = [len(v) for v in voter_pos.values()]

        assert len(cand_dims) > 0, "There must be at least one candidate."
        assert len(set(cand_dims)) == 1, "All candidate positions must have the same number of dimensions."
        assert len(voter_dims) > 0, "There must be at least one voter."
        assert len(set(voter_dims)) == 1, "All voter positions must have the same number of dimensions."
        assert cand_dims[0] == voter_dims[0], "Candidate and voter positions must have the same number of dimensions."

        self.candidates = sorted(list(cand_pos.keys())) 
        self.voters = sorted(list(voter_pos.keys())) 
        self.cand_pos = cand_pos
        self.voter_pos = voter_pos
        self.num_dims = len(list(cand_pos.values())[0]) 
        self.candidate_types = candidate_types or {c:'unknown' for c in self.candidates}

    @property
    def num_cands(self): 
        """
        Returns the number of candidates in the profile.
        """
        return len(self.candidates)

    @property
    def num_voters(self):
        """
        Returns the number of voters in the profile.
        """
        return len(self.voters) 
    
    def voter_position(self, v): 
        """
        Given a voter v, returns their position in the space.
        """
        return self.voter_pos[v]
    
    def candidate_position(self, c):
        """
        Given a candidate c, returns their position in the space.
        """
        return self.cand_pos[c]
    
    def candidate_type(self, c):
        """
        Given a candidate c, returns their type.
        """
        return self.candidate_types[c]

    def set_candidate_types(self, cand_types): 
        """
        Sets the types of each candidate.
        """

        assert set(cand_types.keys()) == set(self.candidates), "The candidate types must be specified for all candidates."

        self.candidate_types = cand_types

    def to_utility_profile(self, 
                           utility_function = None,
                           uncertainty_function=None,
                           batch=False, 
                           return_virtual_cand_positions=False):
        """
        Returns a utility profile corresponding to the spatial profile.  
        
        Args:
            utility_function (callable, optional): A function that takes two vectors and returns a float. The default utility function is the quadratic utility function.
            uncertainty_function (callable, optional): A function that models uncertainty and returns covariance parameters.
            batch (bool, optional): If True, generate positions in batches. Default is False.
            return_virtual_cand_positions (bool, optional): If True, return virtual candidate positions. Default is False.
            
        Returns:    
            UtilityProfile: A utility profile corresponding to the spatial profile.
            (optional) Tuple[UtilityProfile, dict]: The utility profile and virtual candidate positions if `return_virtual_cand_positions` is True.
        """
        import numpy as np
        from pref_voting.generate_spatial_profiles import generate_covariance

        utility_function = utility_function or quadratic_utility

        if uncertainty_function is not None:
            virtual_cand_positions = {}
            for c in self.candidates:
                if batch:
                    covariance = generate_covariance(self.num_dims, *uncertainty_function(self, c, self.voters[0]))
                    positions = np.random.multivariate_normal(self.candidate_position(c), covariance, size=len(self.voters))
                else:
                    positions = [np.random.multivariate_normal(self.candidate_position(c), generate_covariance(self.num_dims, *uncertainty_function(self, c, v))) for v in self.voters]
                virtual_cand_positions[c] = positions
            
            utility_profile = [
                {c: utility_function(self.voter_position(v), virtual_cand_positions[c][vidx]) for c in self.candidates}
                for vidx, v in enumerate(self.voters)
            ]
            
            if return_virtual_cand_positions:
                return UtilityProfile(utility_profile), virtual_cand_positions
            else:
                return UtilityProfile(utility_profile)
        else:
            utility_profile = [
                {c: utility_function(np.array(self.voter_position(v)), np.array(self.candidate_position(c))) for c in self.candidates}
                for v in self.voters
            ]
            return UtilityProfile(utility_profile)
    
    def add_candidate(self, candidate_positions, add_multiple_candidates = False): 
        """
        Add a candidate to the spatial profile. 

        Args: 
            candidate_positions (list): A list of candidate positions
        """

        if add_multiple_candidates: 
            assert all([len(pos) == self.num_dims for pos in candidate_positions]), f"Candidates positions ({candidate_positions}) must be the same dimension as the profile dimension ({self.num_dims})"

            starting_cand_name = self.num_cands

            for c_pos in candidate_positions: 
                self.cand_pos[starting_cand_name] = c_pos
                starting_cand_name += 1

        elif not add_multiple_candidates: 

            assert  len(candidate_positions) == self.num_dims, f"Candidates position ({candidate_positions}) must be the same dimension as the profile dimension ({self.num_dims})"

            starting_cand_name = self.num_cands
            self.cand_pos[starting_cand_name] = candidate_positions

        self.candidates = sorted(list(self.cand_pos.keys())) 

    def move_candidate(self, cand, new_cand_pos): 
        """
        Move cand to a new position
        """

        assert len(new_cand_pos) == self.num_dims, f"The new position {new_cand_pos} must be the same as the profile dimension: {self.num_dims}"

        assert cand in self.candidates, f"Candidate {cand} is not in the profile."

        self.cand_pos[cand] = new_cand_pos


    def to_string(self): 
        """
        Returns a string representation of the spatial profile.
        """

        sp_str = ''
        for c in self.candidates: 
            sp_str += f'C-{c}:{",".join([str(x) for x in self.candidate_position(c)])}_'
        for v in self.voters: 
            sp_str += f'V-{v}:{",".join([str(x) for x in self.voter_position(v)])}_'
        return sp_str[:-1]
    

    @classmethod
    def from_string(cls, sp_str): 
        """
        Returns a spatial profile described by ``sp_str``.

        ``sp_str`` must be in the format produced by the :meth:`pref_voting.SpatialProfile.write` function.
        """

        cand_positions = {}
        voter_positions = {}

        sp_data = sp_str.split('_')

        for d in sp_data: 
            if d.startswith("C-"): 
                cand,positions = d.split(':')
                cand_positions[int(cand[2:])] = np.array([float(x) for x in positions.split(',')])
            elif d.startswith("V-"):
                voter,positions = d.split(':')
                voter_positions[int(voter[2:])] = np.array([float(x) for x in positions.split(',')])

        return cls(cand_positions, voter_positions)

    def view(self, show_cand_labels=False, show_voter_labels=False, bin_width=None, dpi=150):
        """
        Displays the spatial model in a 1D, 2D, or 3D plot.
        
        Args:
            show_cand_labels (optional, bool): If True, displays the labels of each candidate. The default is False.
            show_voter_labels (optional, bool): If True, displays the labels of each voter. The default is False.
                Note: In 1D visualizations, voter labels are disabled regardless of this setting.
            bin_width (optional, float): Width of bins for grouping voters in 1D visualization. If None, a suitable width is calculated.
            dpi (optional, int): Resolution in dots per inch. Default is 150.
        """
        assert self.num_dims <= 3, "Can only view profiles with 1, 2, or 3 dimensions"
        sns.set_theme(style="darkgrid")
        
        # Define the candidate color consistently across all dimensions
        candidate_color = "red"
        
        if self.num_dims == 1:
            # Get all voter positions
            voter_positions = [self.voter_position(v)[0] for v in self.voters]
            
            # Calculate histogram data
            if bin_width is None:
                # Auto-calculate a reasonable bin width based on data range
                position_range = max(voter_positions) - min(voter_positions)
                bin_width = max(position_range / 20, 0.05)
            
            # Create histogram data - exact counts
            bins = {}
            for pos in voter_positions:
                # Round to nearest bin
                binned_pos = round(pos / bin_width) * bin_width
                bins[binned_pos] = bins.get(binned_pos, 0) + 1
            
            # Get the bin positions and counts
            bin_positions = list(bins.keys())
            bin_counts = list(bins.values())
            max_count = max(bin_counts) if bin_counts else 1
            
            # Create figure with sufficient space for labels and high resolution
            fig, ax = plt.subplots(figsize=(10, 6), dpi=dpi)
            
            # Calculate the space needed for candidates at the top
            candidate_area_height = max_count * 0.3
            
            # Plot the bars for voters
            bars = ax.bar(
                bin_positions,
                bin_counts,
                width=bin_width*0.8, 
                alpha=0.6,
                color="blue",
                label="Voters"
            )
            
            # Calculate position for candidates at the top
            top_line_y = max_count + 0.2
            candidate_y_pos = top_line_y + candidate_area_height * 0.3
            candidate_positions = [self.candidate_position(c)[0] for c in self.candidates]
            
            # Plot candidates above the histogram
            cand_scatter = ax.scatter(candidate_positions, [candidate_y_pos] * len(self.candidates), 
                    color=candidate_color, marker='X', s=100, zorder=5)
            
            # Create a custom legend box
            legend = ax.legend([cand_scatter, bars], ['Candidates', 'Voters'], 
                            loc='upper right', 
                            bbox_to_anchor=(0.99, 0.99))
            
            # Draw the figure to get legend position for label placement
            fig.canvas.draw()
            
            # Get legend position for detecting label overlap
            if legend:
                legend_bbox = legend.get_window_extent().transformed(ax.transData.inverted())
                legend_left = legend_bbox.x0
                legend_right = legend_bbox.x1
                legend_width = legend_right - legend_left
            else:
                legend_left = float('inf')
                legend_right = float('inf')
                legend_width = 0
            
            if show_cand_labels:
                # Add labels to each candidate
                for c in self.candidates:
                    pos = self.candidate_position(c)[0]
                    
                    # Buffer to detect legend overlap
                    legend_buffer = 0.05 * legend_width
                    
                    # Check if candidate is under or near the legend
                    if (pos >= legend_left - legend_buffer) and (pos <= legend_right + legend_buffer):
                        # Place label below the marker to avoid legend overlap
                        ax.annotate(c, (pos, candidate_y_pos), xytext=(0, -25), 
                                textcoords='offset points', ha='center', va='top',
                                fontsize=13, fontweight='bold', color=candidate_color)
                    else:
                        # Otherwise place it above
                        ax.annotate(c, (pos, candidate_y_pos), xytext=(0, 15), 
                                textcoords='offset points', ha='center', va='bottom',
                                fontsize=13, fontweight='bold', color=candidate_color)
            
            # Set axis labels
            ax.set_xlabel('Position')
            ax.set_ylabel('Number of voters')
            
            # Configure y-axis to show integers for the histogram area
            ax.yaxis.set_major_locator(plt.MaxNLocator(integer=True))
            
            # Set y-limits to include the candidate area
            ax.set_ylim(0, top_line_y + candidate_area_height)
            
            # Hide y-ticks in the candidate area
            yticks = [t for t in ax.get_yticks() if t <= max_count]
            ax.set_yticks(yticks)
            
            # Adjust the figure layout
            plt.tight_layout(rect=[0, 0, 1, 0.97])
            
            plt.show()
        
        elif self.num_dims == 2:
    
            fig, ax = plt.subplots(figsize=(10, 6), dpi=dpi)
            
            # Get voter positions
            x_voters = [self.voter_position(v)[0] for v in self.voters]
            y_voters = [self.voter_position(v)[1] for v in self.voters]
            
            # Plot voters with semi-transparency to visualize density
            voter_scatter = ax.scatter(x_voters, y_voters, 
                                    color="blue", alpha=0.2,
                                    edgecolor="black", linewidth=0.3,
                                    s=30,
                                    label="Voters")
            
            # Plot candidates
            x_cand = [self.candidate_position(c)[0] for c in self.candidates]
            y_cand = [self.candidate_position(c)[1] for c in self.candidates]
            cand_scatter = ax.scatter(x_cand, y_cand, 
                                    color=candidate_color, marker='X', s=100,
                                    edgecolor="white", linewidth=0.7,
                                    zorder=5,
                                    label="Candidates")
            
            # Create a legend
            ax.legend([cand_scatter, voter_scatter], ['Candidates', 'Voters'], loc='upper right')
            
            if show_cand_labels:
                for c in self.candidates:
                    pos = self.candidate_position(c)
                    text = ax.annotate(c, (pos[0], pos[1]), xytext=(0, 10), 
                            textcoords='offset points', ha='center', va='bottom',
                            fontsize=11, fontweight='bold', color=candidate_color)
                    
                    # Add white outline to text for better visibility
                    text.set_path_effects([
                        path_effects.Stroke(linewidth=1.5, foreground='white'),
                        path_effects.Normal()
                    ])
                    
            if show_voter_labels:
                for v in self.voters:
                    pos = self.voter_position(v)
                    ax.annotate(v + 1, (pos[0], pos[1]), fontsize=8)
                    
            ax.set_xlabel('Dimension 1')
            ax.set_ylabel('Dimension 2')
            plt.tight_layout()
            plt.show()
        
        elif self.num_dims == 3:

            fig = plt.figure(figsize=(10, 6), dpi=dpi)
            ax = fig.add_subplot(111, projection='3d')
            
            # Fetch all positions
            x_voters = [self.voter_position(v)[0] for v in self.voters]
            y_voters = [self.voter_position(v)[1] for v in self.voters]
            z_voters = [self.voter_position(v)[2] for v in self.voters]
            
            x_cand = [self.candidate_position(c)[0] for c in self.candidates]
            y_cand = [self.candidate_position(c)[1] for c in self.candidates]
            z_cand = [self.candidate_position(c)[2] for c in self.candidates]
            
            # Plot voters with high transparency for better visibility through clusters
            voter_scatter = ax.scatter(x_voters, y_voters, z_voters, 
                                    color="blue", alpha=0.1,
                                    edgecolor="black", linewidth=0.5,
                                    s=30,
                                    label="Voters")
            
            # Plot candidate markers with white outline for better visibility
            cand_scatter = ax.scatter(x_cand, y_cand, z_cand, 
                                    color=candidate_color, marker="X", s=40,
                                    edgecolor="white", linewidth=0.7,
                                    label="Candidates")
            
            # Add voter labels if requested
            if show_voter_labels:
                for v in self.voters:
                    pos = self.voter_position(v)
                    ax.text(pos[0], pos[1], pos[2], str(v + 1), fontsize=8)
            
            # Add candidate labels
            if show_cand_labels:
                for c in self.candidates:
                    pos = self.candidate_position(c)
                    text = ax.text(pos[0], pos[1], pos[2] + 0.15, c, 
                        fontsize=11, fontweight='bold', color=candidate_color,
                        ha='center', va='bottom')
                    
                    # Add white outline to text for better visibility
                    text.set_path_effects([
                        path_effects.Stroke(linewidth=1.5, foreground='white'),
                        path_effects.Normal()
                    ])
            
            # Create legend
            ax.legend([cand_scatter, voter_scatter], ['Candidates', 'Voters'], loc='upper right')
            
            # Set axis labels
            ax.set_xlabel('Dimension 1')
            ax.set_ylabel('Dimension 2')
            ax.set_zlabel('Dimension 3')
            
            plt.tight_layout()
            plt.show()
    
    def display(self): 
        """
        Displays the positions of each candidate and voter in the profile.

        """
        print("Candidates: ")
        for c in self.candidates: 
            print("Candidate ", c, " position: ", self.candidate_position(c))

        print("\nVoters: ")
        for v in self.voters:
            print("Voter ", v, " position: ", self.voter_position(v))


from pref_voting.dominance_axioms import *
from pref_voting.monotonicity_axioms import *
from pref_voting.invariance_axioms import *
from pref_voting.strategic_axioms import *
from pref_voting.variable_voter_axioms import *
from pref_voting.variable_candidate_axioms import *
from pref_voting.axiom import Axiom


axioms_dict = {name: obj for name, obj in globals().items() if isinstance(obj, Axiom)}



"""
    File: weighted_majority_graphs.py
    Author: Wes Holliday (wesholliday@berkeley.edu) and Eric Pacuit (epacuit@umd.edu)
    Date: January 5, 2022
    Updated: July 12, 2022
    Updated: December 19, 2022
    
    Majority Graphs, Margin Graphs and Support Graphs
"""

import networkx as nx
from tabulate import tabulate
import matplotlib.pyplot as plt
import string
from itertools import combinations, permutations
from ortools.linear_solver import pywraplp

import numpy as np

class MajorityGraph(object):
    """A majority graph is an asymmetric directed graph.  The nodes are the candidates and an edge from candidate :math:`c` to :math:`d` means that :math:`c` is majority preferred to :math:`d`.

    :param candidates: List of the candidates.  To be used as nodes in the majority graph.
    :type candidates: list[int] or  list[str]
    :param edges: List of the pairs of candidates describing the edges in the majority graph.   If :math:`(c,d)` is in the list of edges, then there is an edge from :math:`c` to :math:`d`.
    :type edges: list
    :param cmap: Dictionary mapping candidates to candidate names (strings).  If not provided, each candidate name is mapped to itself.
    :type cmap: dict[int: str], optional

    :Example:

    The following code creates a majority graph in which 0 is majority preferred to 1, 1 is majority preferred to 2, and 2 is majority preferred to 0:

    .. code-block:: python

            mg = MajorityGraph([0, 1, 2], [(0,1), (1,2), (2,0)])

    .. warning:: Currently, there is no check that the edge relation is asymmetric.  It is assumed that the user provides an appropriate set of edges.
    """

    def __init__(self, candidates, edges, cmap=None):
        """constructer method"""

        mg = nx.DiGraph()
        mg.add_nodes_from(candidates)
        mg.add_edges_from(edges)
        self.mg = mg
        """A networkx DiGraph object representing the majority graph."""

        self.cmap = cmap if cmap is not None else {c: str(c) for c in candidates}
        
        self.candidates = list(candidates)
        """The list of candidates."""
        
        self.num_cands = len(self.candidates)
        """The number of candidates."""

        self.cindices = list(range(self.num_cands))
        self._cand_to_cindex = {c: i for i, c in enumerate(self.candidates)}
        self.cand_to_cindex = lambda c: self._cand_to_cindex[c]
        self._cindex_to_cand = {i: c for i, c in enumerate(self.candidates)}
        self.cindex_to_cand = lambda i: self._cindex_to_cand[i]
        """A dictionary mapping each candidate to its index in the list of candidates and vice versa."""

        self.maj_matrix = [[False for c2 in self.cindices] for c1 in self.cindices]
        """A matrix of Boolean values representing the majority graph."""

        for c1_idx in self.cindices:
            for c2_idx in self.cindices:
                if mg.has_edge(self.cindex_to_cand(c1_idx), self.cindex_to_cand(c2_idx)):
                    self.maj_matrix[c1_idx][c2_idx] = True
                    self.maj_matrix[c2_idx][c1_idx] = False
                elif mg.has_edge(self.cindex_to_cand(c2_idx), self.cindex_to_cand(c1_idx)):
                    self.maj_matrix[c2_idx][c1_idx] = True
                    self.maj_matrix[c1_idx][c2_idx] = False

    def margin(self, c1, c2):
        raise Exception("margin is not implemented for majority graphs.")

    def support(self, c1, c2):
        raise Exception("support is not implemented for majority graphs.")

    def ratio(self, c1, c2):
        raise Exception("ratio is not implemented for majority graphs.")

    @property
    def edges(self):
        """Returns a list of the edges in the majority graph."""

        return list(self.mg.edges)

    @property
    def is_tournament(self):
        """Returns True if the majority graph is a **tournament** (there is an edge between any two distinct nodes)."""

        return all([
            self.mg.has_edge(c1, c2) or self.mg.has_edge(c2, c1)
            for c1 in self.candidates
            for c2 in self.candidates
            if c1 != c2
        ])

    def majority_prefers(self, c1, c2):
        """Returns true if there is an edge from `c1` to `c2`."""
        return self.mg.has_edge(c1, c2)

    def is_tied(self, c1, c2):
        """Returns true if there is no edge from `c1` to `c2` or from `c2` to `c1`."""
        return not self.mg.has_edge(c1, c2) and not self.mg.has_edge(c2, c1)

    def copeland_scores(self, curr_cands=None, scores=(1, 0, -1)):
        """The Copeland scores in the profile restricted to the candidates in ``curr_cands``.

        The **Copeland score** for candidate :math:`c` is calculated as follows:  :math:`c` receives ``scores[0]`` points for every candidate that  :math:`c` is majority preferred to, ``scores[1]`` points for every candidate that is tied with :math:`c`, and ``scores[2]`` points for every candidate that is majority preferred to :math:`c`. The default ``scores`` is ``(1, 0, -1)``.

        :param curr_cands: restrict attention to candidates in this list. Defaults to all candidates in the profile if not provided.
        :type curr_cands: list[int], optional
        :param scores: the scores used to calculate the Copeland score of a candidate :math:`c`: ``scores[0]`` is for the candidates that :math:`c` is majority preferred to; ``scores[1]`` is for the candidates tied with :math:`c`; and ``scores[2]`` is for the candidates majority preferred to :math:`c`.  The default value is ``scores = (1, 0, -1)``
        :type scores: tuple[int], optional
        :returns: a dictionary associating each candidate in ``curr_cands`` with its Copeland score.

        """

        wscore, tscore, lscore = scores
        candidates = self.candidates if curr_cands is None else curr_cands
        c_scores = {c: 0.0 for c in candidates}
        for c1 in candidates:
            for c2 in candidates:
                if self.majority_prefers(c1, c2):
                    c_scores[c1] += wscore
                elif self.majority_prefers(c2, c1):
                    c_scores[c1] += lscore
                elif c1 != c2:
                    c_scores[c1] += tscore
        return c_scores

    def dominators(self, cand, curr_cands=None):
        """Returns the list of candidates that are majority preferred to ``cand`` in the majority graph restricted to ``curr_cands``."""
        candidates = self.candidates if curr_cands is None else curr_cands

        return [c for c in candidates if self.majority_prefers(c, cand)]

    def dominates(self, cand, curr_cands=None):
        """Returns the list of candidates that ``cand`` is majority preferred to in the majority graph restricted to ``curr_cands``."""
        candidates = self.candidates if curr_cands is None else curr_cands

        return [c for c in candidates if self.majority_prefers(cand, c)]

    def condorcet_winner(self, curr_cands=None):
        """Returns the Condorcet winner in the profile restricted to ``curr_cands`` if one exists, otherwise returns None.

        The **Condorcet winner** is the candidate that is majority preferred to every other candidate.
        """

        curr_cands = curr_cands if curr_cands is not None else self.candidates

        cw = None
        for c1 in curr_cands:
            if all([self.majority_prefers(c1, c2) for c2 in curr_cands if c1 != c2]):
                cw = c1
                break  # if a Condorcet winner exists, then it is unique
        return cw

    def weak_condorcet_winner(self, curr_cands=None):
        """Returns a list of the weak Condorcet winners in the profile restricted to ``curr_cands`` (which may be empty).

        A candidate :math:`c` is a  **weak Condorcet winner** if there is no other candidate that is majority preferred to :math:`c`.

        .. note:: While the Condorcet winner is unique if it exists, there may be multiple weak Condorcet winners.
        """

        curr_cands = curr_cands if curr_cands is not None else self.candidates

        weak_cw = list()
        for c1 in curr_cands:
            if not any(
                [self.majority_prefers(c2, c1) for c2 in curr_cands if c1 != c2]
            ):
                weak_cw.append(c1)
        return sorted(weak_cw) if len(weak_cw) > 0 else None

    def condorcet_loser(self, curr_cands=None):
        """Returns the Condorcet loser in the profile restricted to ``curr_cands`` if one exists, otherwise returns None.

        A candidate :math:`c` is a  **Condorcet loser** if every other candidate  is majority preferred to :math:`c`.
        """

        curr_cands = curr_cands if curr_cands is not None else self.candidates

        cl = None
        for c1 in curr_cands:
            if all([self.majority_prefers(c2, c1) for c2 in curr_cands if c1 != c2]):
                cl = c1
                break  # if a Condorcet loser exists, then it is unique
        return cl

    def cycles(self, curr_cands = None):
        """Returns True if the margin graph has a cycle.

        This uses the networkx method ``networkx.find_cycle`` to find the cycles in ``self.mg``.

        :Example:

        .. exec_code::

            from pref_voting.weighted_majority_graphs import MajorityGraph
            mg = MajorityGraph([0,1,2], [(0,1), (1,2), (0,2)])
            print(f"The cycles in the majority graph are {mg.cycles()}")
            mg = MajorityGraph([0,1,2], [(0,1), (1,2), (2,0)])
            print(f"The cycles in the majority graph are {mg.cycles()}")
            mg = MajorityGraph([0,1,2,3], [(0,1), (3,0), (1,2), (3,1), (2,0), (3,2)])
            print(f"The cycles in the majority graph are {mg.cycles()}")

        """

        if curr_cands is None:
            return list(nx.simple_cycles(self.mg))
        else:
            mg = nx.DiGraph()
            subgraph = mg.subgraph(curr_cands).copy()
            return list(nx.simple_cycles(subgraph))


    def has_cycle(self, curr_cands = None):
        """Returns True if there is a cycle in the majority graph."""
        if curr_cands is None:
            try:
                cycle = nx.find_cycle(self.mg)
            except:
                cycle = list()
        else:
            mg = nx.DiGraph()
            subgraph = mg.subgraph(curr_cands).copy()
            try:
                cycle = nx.find_cycle(subgraph)
            except:
                cycle = list()

        return len(cycle) != 0

    def remove_candidates(self, cands_to_ignore):
        """Remove all candidates from ``cands_to_ignore`` from the Majority Graph.

        :param cands_to_ignore: list of candidates to remove from the profile
        :type cands_to_ignore: list[int]
        :returns: a majority graph with candidates from ``cands_to_ignore`` removed and a dictionary mapping the candidates from the new profile to the original candidate names.

        :Example:

        .. exec_code::

            from pref_voting.weighted_majority_graphs import MajorityGraph
            mg = MajorityGraph([0, 1, 2], [(0, 1), (1, 2), (2, 0)])
            print(f"Candidates: {mg.candidates}")
            print(f"Edges: {mg.edges}")
            mg_new = mg.remove_candidates([1])
            print(f"Candidates: {mg_new.candidates}")
            print(f"Edges: {mg_new.edges}")
        """

        new_cands = [c for c in self.candidates if c not in cands_to_ignore]

        new_edges = [e for e in self.edges if e[0] in new_cands and e[1] in new_cands]

        new_cmap = {c: cname for c, cname in self.cmap.items() if c in new_cands}

        return MajorityGraph(new_cands, new_edges, cmap=new_cmap)

    def to_networkx(self): 
        """
        Return a networkx weighted DiGraph representing the margin graph. 
        """

        return self.mg

    def description(self): 
        """
        Returns a string describing the Majority Graph.
        """
        return f"MajorityGraph({self.candidates}, {self.edges}, cmap={self.cmap})"

    def display(self, cmap=None, curr_cands=None):
        """Display a majority graph (restricted to ``curr_cands``) using networkx.draw.

        :param cmap: the candidate map to use (overrides the cmap associated with this majority graph)
        :type cmap: dict[int,str], optional
        :param curr_cands: list of candidates
        :type curr_cands: list[int], optional
        :rtype: None

        :Example:

        .. code::

            from pref_voting.weighted_majority_graphs import MajorityGraph
            mg = MajorityGraph([0,1,2], [(0,1), (1,2), (2,0)])
            mg.display()

        .. image:: ./maj_graph_ex1.png
            :width: 400
            :alt: Alternative text

        """

        cmap = cmap if cmap is not None else self.cmap
        curr_cands = self.candidates if curr_cands is None else curr_cands

        mg = nx.DiGraph()
        mg.add_nodes_from([cmap[c] for c in curr_cands])
        mg.add_edges_from(
            [
                (cmap[c1], cmap[c2])
                for c1, c2 in self.mg.edges
                if c1 in curr_cands and c2 in curr_cands
            ]
        )

        pos = nx.circular_layout(mg)

        nx.draw(
            mg,
            pos,
            font_size=20,
            font_color="white",
            node_size=700,
            width=1.5,
            with_labels=True,
        )
        plt.show()

    def display_cycles(self, cmap=None):
        """
        Display the cycles in the margin graph.

        Args:
            cmap (dict, optional): The cmap used to map candidates to candidate names
        """

        cycles = self.cycles()
        
        print(f"There {'are' if len(cycles) != 1 else 'is'} {len(cycles)} {'cycle' if len(cycles) == 1 else 'cycles'}{':' if len(cycles) > 0 else '.'} \n")
        for cycle in cycles: 
            cmap = cmap if cmap is not None else self.cmap
            cmap_inverse = {cname: c for c, cname in cmap.items()}
            mg_with_cycle = nx.DiGraph()

            mg_with_cycle.add_nodes_from([cmap[c] for c in self.candidates])
            mg_with_cycle.add_edges_from([(cmap[e[0]], cmap[e[1]]) for e in self.edges])

            cands = self.candidates
            mg_edges = list(self.edges)

            cycle_edges = [(cmap[c], cmap[cycle[cidx + 1]]) for cidx,c in enumerate(cycle[0:-1])] + [(cmap[cycle[-1]], cmap[cycle[0]])]

            cands_in_cycle = [cmap[c1] for c1 in cands if c1 in cycle]

            node_colors = ["blue" if n in cands_in_cycle else "lightgray" for n in mg_with_cycle.nodes ]

            pos = nx.circular_layout(mg_with_cycle)
            nx.draw(mg_with_cycle, pos, width=1.5, edge_color="white")
            nx.draw_networkx_nodes(
                mg_with_cycle, pos, node_color=node_colors, node_size=700
            )
            nx.draw_networkx_labels(mg_with_cycle, pos, font_size=20, font_color="white")
            nx.draw_networkx_edges(
                mg_with_cycle,
                pos,
                edgelist=cycle_edges,
                width=10,
                alpha=1.0,
                edge_color="b",
                arrowsize=25,
                min_target_margin=15,
                node_size=700,
            )

            nx.draw_networkx_edges(
                mg_with_cycle,
                pos,
                edgelist=[(cmap[e[0]], cmap[e[1]]) for e in mg_edges if (cmap[e[0]], cmap[e[1]]) not in cycle_edges],
                width=1.5,
                 edge_color="lightgray",
                arrowsize=15,
                min_target_margin=15,
            )
            
            ax = plt.gca()
            ax.set_frame_on(False)
            plt.show()
            
    def to_latex(self, cmap=None, new_cand=None):
        """Outputs TikZ code for displaying the majority graph.

        :param cmap: the candidate map to use (overrides the cmap associated with this majority graph)
        :type cmap: dict[int,str], optional
        :param new_cand: the candidate that is displayed on the far right,  *only used for displaying 5 candidates*.
        :type new_cand: int
        :rtype: str

        .. warning:: This works best for 3, 4 or 5 candidates.   It will produce the code for more than 5 outputs, but the positioning of the nodes may need to be modified.

        :Example:

        .. exec_code::

            from pref_voting.weighted_majority_graphs import MajorityGraph
            mg = MajorityGraph([0,1,2], [(0,1), (1,2), (2,0)])
            print(mg.to_latex())
            print(mg.to_latex(cmap = {0:"a", 1:"b", 2:"c"}))

        """
        if len(self.candidates) == 3:
            return three_cand_tikz_str(self, cmap=cmap)
        elif len(self.candidates) == 4:
            return four_cand_tikz_str(self, cmap=cmap)
        elif len(self.candidates) == 5:
            return five_cand_tikz_str(self, cmap=cmap, new_cand=new_cand)
        else:
            pos = nx.circular_layout(self.mg)
            return to_tikz_str(self, pos, cmap=cmap)

    @classmethod
    def from_profile(cls, profile, cmap=None):
        """Generates a majority graph from a :class:`Profile`.

        :param profile: the profile
        :type profile: Profile
        :param cmap: the candidate map to use (overrides the cmap associated with this majority graph)
        :type cmap: dict[int,str], optional
        :rtype: str

        :Example:

        .. exec_code::

            from pref_voting.profiles import Profile
            from pref_voting.weighted_majority_graphs import MajorityGraph
            prof = Profile([[0,1,2], [1,2,0], [2,0,1]])
            mg = MajorityGraph.from_profile(prof)
            print(mg.edges)

            # it is better to use the Profile method
            mg = prof.majority_graph()
            print(mg.edges)

        """
        cmap = profile.cmap if cmap is None else cmap
        return cls(
            profile.candidates,
            [
                (c1, c2)
                for c1 in profile.candidates
                for c2 in profile.candidates
                if profile.majority_prefers(c1, c2)
            ],
            cmap=cmap,
        )

    def __add__(self, other_mg):
        """
        Add to majority graphs together.  The result is a majority graph with the union of the candidates and edges of the two majority graphs.  If there is an edge from :math:`c` to :math:`d` in both majority graphs, then the edge is included in the resulting majority graph.  If there is an edge from :math:`c` to :math:`d` in neither majority graph, then there is no edge from :math:`c` to :math:`d` in the resulting majority graph.
        """
        new_cands = list(set(self.candidates + other_mg.candidates))

        _new_edges = list(set(self.edges + other_mg.edges))

        new_edges = list()
        for c1, c2 in _new_edges:
            if (self.majority_prefers(c1, c2) and other_mg.majority_prefers(c2, c1)) or (self.majority_prefers(c2, c1) and other_mg.majority_prefers(c1, c2)):
                continue
            else: 
                new_edges.append((c1, c2))
        return MajorityGraph(new_cands, new_edges)
    
    def __eq__(self, other_mg): 
        """
        Check if two majority graphs are equal. 
        """
        return self.candidates == other_mg.candidates and sorted(self.edges) == sorted(other_mg.edges)

class MarginGraph(MajorityGraph):
    """A margin graph is a weighted asymmetric directed graph.  The nodes are the candidates and an edge from candidate :math:`c` to :math:`d` with weight :math:`w` means that :math:`c` is majority preferred to :math:`d` by a **margin** of :math:`w`.

    :param candidates: List of the candidates.  To be used as nodes in the majority graph.
    :type candidates: list[int] or  list[str]
    :param w_edges: List of the pairs of candidates describing the edges in the majority graph.   If :math:`(c,d,w)` is in the list of edges, then there is an edge from :math:`c` to :math:`d` with weight :math:`w`.
    :type w_edges: list
    :param cmap: Dictionary mapping candidates to candidate names (strings).  If not provided, each candidate name is mapped to itself.
    :type cmap: dict[int: str], optional

    :Example:

    The following code creates a margin graph in which 0 is majority preferred to 1 by a margin of 1, 1 is majority preferred to 2 by a margin of 3, and 2 is majority preferred to 0 by a margin of 5:

    .. code-block:: python

        mg = MarginGraph([0, 1, 2], [(0,1,1), (1,2,3), (2,0,5)])

    .. warning:: Currently, there is no check that the edge relation is asymmetric or that weights of edges are positive.  It is assumed that the user provides an appropriate set of edges with weights.
    """

    def __init__(self, candidates, w_edges, cmap=None):
        """constructor method"""

        super().__init__(candidates, [(e[0], e[1]) for e in w_edges], cmap=cmap)

        self.margin_matrix = [[0 for c2 in self.cindices] for c1 in self.cindices]
        """The margin matrix, where the :math:`(i, j)`-entry is the number of voters who rank candidate with index :math:`i` above the candidate with index :math:`j` minus the number of voters  who rank candidate with index :math:`j` above the candidate with index :math:`i`. """

        for c1, c2, margin in w_edges:
            self.margin_matrix[self.cand_to_cindex(c1)][self.cand_to_cindex(c2)] = margin
            self.margin_matrix[self.cand_to_cindex(c2)][self.cand_to_cindex(c1)] = -1 * margin

    def margin(self, c1, c2):
        """Returns the margin of ``c1`` over ``c2``."""
        return self.margin_matrix[self.cand_to_cindex(c1)][self.cand_to_cindex(c2)]

    def strength_matrix(self, curr_cands = None, strength_function = None): 
        """
        Return the strength matrix of the profile.  The strength matrix is a matrix where the entry in row :math:`i` and column :math:`j` is the number of voters that rank the candidate with index :math:`i` over the candidate with index :math:`j`.  If ``curr_cands`` is provided, then the strength matrix is restricted to the candidates in ``curr_cands``.  If ``strength_function`` is provided, then the strength matrix is computed using the strength function."""
        
        if curr_cands is not None: 
            cindices = [cidx for cidx, _ in enumerate(curr_cands)]
            cindex_to_cand = lambda cidx: curr_cands[cidx]
            cand_to_cindex = lambda c: cindices[curr_cands.index(c)]
            strength_function = self.margin if strength_function is None else strength_function
            strength_matrix = np.array([[strength_function(cindex_to_cand(a_idx), cindex_to_cand(b_idx)) for b_idx in cindices] for a_idx in cindices])
        else:  
            cindices = self.cindices
            cindex_to_cand = self.cindex_to_cand
            cand_to_cindex = self.cand_to_cindex
            strength_matrix = np.array(self.margin_matrix) if strength_function is None else np.array([[strength_function(cindex_to_cand(a_idx), cindex_to_cand(b_idx)) for b_idx in cindices] for a_idx in cindices])

        return strength_matrix, cand_to_cindex

    @property
    def edges(self):
        """Returns a list of the weighted edges in the margin graph."""

        return [(c1, c2, self.margin(c1, c2)) for c1, c2 in self.mg.edges]

    def remove_candidates(self, cands_to_ignore):
        """Remove all candidates from ``cands_to_ignore`` from the Majority Graph.

        :param cands_to_ignore: list of candidates to remove from the profile
        :type cands_to_ignore: list[int]
        :returns: a majority graph with candidates from ``cands_to_ignore`` removed and a dictionary mapping the candidates from the new profile to the original candidate names.

        :Example:

        .. exec_code::

            from pref_voting.weighted_majority_graphs import MarginGraph
            mg = MarginGraph([0, 1, 2], [(0, 1, 11), (1, 2, 13), (2, 0, 5)])
            print(f"Candidates: {mg.candidates}")
            print(f"Edges: {mg.edges}")
            mg_new = mg.remove_candidates([1])
            print(f"Candidates: {mg_new.candidates}")
            print(f"Edges: {mg_new.edges}")
        """

        new_cands = [c for c in self.candidates if c not in cands_to_ignore]

        new_edges = [e for e in self.edges if e[0] in new_cands and e[1] in new_cands]

        new_cmap = {c: cname for c, cname in self.cmap.items() if c in new_cands}

        return MarginGraph(new_cands, new_edges, cmap=new_cmap)

    def majority_prefers(self, c1, c2):
        """Returns True if the margin of ``c1`` over ``c2`` is positive."""
        return self.margin_matrix[self.cand_to_cindex(c1)][self.cand_to_cindex(c2)] > 0

    def is_tied(self, c1, c2):
        """Returns True if the margin ``c1`` over ``c2`` is zero."""
        return self.margin_matrix[self.cand_to_cindex(c1)][self.cand_to_cindex(c2)] == 0

    def is_uniquely_weighted(self):
        """Returns True if all the margins between distinct candidates are unique and there is no 0 margin between distinct candidates."""
        has_zero_margins = any(
            [
                self.margin(c1, c2) == 0
                for c1 in self.candidates
                for c2 in self.candidates
                if c1 != c2
            ]
        )
        return not has_zero_margins and len(
            list(set([self.margin(e[0], e[1]) for e in self.mg.edges]))
        ) == len(self.mg.edges)


    def to_networkx(self): 
        """
        Return a networkx weighted DiGraph representing the margin graph. 
        """

        g = nx.DiGraph()
        g.add_nodes_from(self.candidates)
        g.add_weighted_edges_from(self.edges)

        return g

    def debord_profile(self): 
        """
        Find a profile that generates the margin graph using the algorithm from Debord's (1987) proof.
        """
        
        from pref_voting.profiles import Profile
    
        candidates = self.candidates

        ranks = list()
        rcounts = list()

        if all([w % 2 == 0 for _,_,w in self.edges]):
            for c1, c2, w in self.edges:
                other_cands = [c for c in candidates if c != c1 and c != c2]

                lin_ord1 = sorted(other_cands)
                lin_ord2 = sorted(other_cands, reverse=True)

                ranks.append([c1, c2] + lin_ord1)
                rcounts.append(w//2)
                ranks.append(lin_ord2 + [c1, c2])
                rcounts.append(w//2)
            return Profile(ranks, rcounts, cmap=self.cmap)
        
        elif all([w % 2 == 1 for _,_,w in self.edges]): # all weights are odd
            single_prof = Profile([sorted(candidates)], [1])

            for c1, c2, w in self.edges:
                other_cands = [c for c in candidates if c != c1 and c != c2]
                lin_ord1 = sorted(other_cands)
                lin_ord2 = sorted(other_cands, reverse=True)
                if w-single_prof.margin(c1, c2) > 0:
                    ranks.append([c1, c2] + lin_ord1)
                    rcounts.append((w-single_prof.margin(c1, c2))//2)
                    ranks.append(lin_ord2 + [c1, c2])
                    rcounts.append((w-single_prof.margin(c1, c2))//2)
            ranks.append(sorted(candidates))
            rcounts.append(1)
            return Profile(ranks, rcounts, cmap=self.cmap)
        else: 
            print("Cannot find a Profile since the weights do not all have the same parity.")
            return None
        
    def minimal_profile(self): 
        """
        Use an integer linear program to find a minimal profile generating the margin graph. 
        """
        from pref_voting.profiles import Profile

        solver = pywraplp.Solver.CreateSolver("SAT")

        num_cands = len(self.candidates)
        rankings = list(permutations(range(num_cands)))
        
        ranking_to_var = dict()
        infinity = solver.infinity()
        for ridx, r in enumerate(rankings): 
            _v = solver.IntVar(0.0, infinity, f"x{ridx}")
            ranking_to_var[r] = _v

        nv = solver.IntVar(0.0, infinity, "nv")
        equations = list()
        for c1 in self.candidates: 
            for c2 in self.candidates: 
                if c1 != c2: 
                    margin = self.margin(c1, c2)
                    if margin >= 0: 
                        rankings_c1_over_c2 = [ranking_to_var[r] for r in rankings if r.index(c1) < r.index(c2)]
                        rankings_c2_over_c1 = [ranking_to_var[r] for r in rankings if r.index(c2) < r.index(c1)]
                        equations.append(sum(rankings_c1_over_c2) == margin + sum(rankings_c2_over_c1))
                        
        equations.append(nv == sum(list(ranking_to_var.values())))

        for eq in equations: 
            solver.Add(eq)

        solver.Minimize(nv)

        status = solver.Solve()

        if status == pywraplp.Solver.INFEASIBLE:
            print("Error: Did not find a solution.")
            return None
        
        if status != pywraplp.Solver.OPTIMAL: 
            print("Warning: Did not find an optimal solution.")

        _ranks = list()
        _rcounts = list()

        for r,v in ranking_to_var.items(): 

            if v.solution_value() > 0: 
                _ranks.append(r)
                _rcounts.append(int(v.solution_value()))
                if not v.solution_value().is_integer(): 
                    print("ERROR: Found non integer, ", v.solution_value())
                    return None
        return Profile(_ranks, rcounts = _rcounts)

    def normalize_ordered_weights(self):
        """
        Returns a MarginGraph with the same order of the edges, except that the weights are 2, 4, 6,...

        .. important::

            This function returns a margin graph that has the same ordering of the edges, but the edges may have different weights.  Qualitative margin graph invariant 
            voting methods will identify the same winning sets for both graphs. 

        """

        sorted_edges = sorted(self.edges, key=lambda e: e[2])
        sorted_margins = sorted(
            [
                self.margin(c1, c2)
                for c1 in self.candidates
                for c2 in self.candidates
                if self.margin(c1, c2) > 0
            ]
        )
        curr_margin = sorted_margins[0]
        new_margin = 2
        new_edges = list()
        for e in sorted_edges:
            if e[2] > curr_margin:
                curr_margin = e[2]
                new_margin += 2
                new_edges.append((e[0], e[1], new_margin))

            else:
                new_edges.append((e[0], e[1], new_margin))

        return MarginGraph(self.candidates, new_edges, cmap=self.cmap)

    def description(self): 
        """
        Returns a string describing the Margin Graph.
        """
        return f"MarginGraph({self.candidates}, {self.edges}, cmap={self.cmap})"

    def display(self, curr_cands=None, cmap=None):
        """Display a margin graph (restricted to ``curr_cands``) using networkx.draw.

        :param cmap: the candidate map to use (overrides the cmap associated with this majority graph)
        :type cmap: dict[int,str], optional
        :param curr_cands: list of candidates
        :type curr_cands: list[int], optional
        :rtype: None

        :Example:

        .. code::

            from pref_voting.weighted_majority_graphs import MarginGraph
            mg = MarginGraph([0,1,2], [(0,1,3), (1,2,1), (2,0,5)])
            mg.display()

        .. image:: ./margin_graph_ex1.png
            :width: 400
            :alt: Alternative text

        """

        cmap = cmap if cmap is not None else self.cmap
        curr_cands = self.candidates if curr_cands is None else curr_cands

        mg = nx.DiGraph()
        mg.add_nodes_from([cmap[c] for c in curr_cands])
        mg.add_weighted_edges_from(
            [
                (cmap[c1], cmap[c2], self.margin(c1, c2))
                for c1, c2 in self.mg.edges
                if c1 in curr_cands and c2 in curr_cands
            ]
        )

        pos = nx.circular_layout(mg)

        nx.draw(
            mg,
            pos,
            font_size=20,
            font_color="white",
            node_size=700,
            width=1.5,
            with_labels=True,
        )
        labels = nx.get_edge_attributes(mg, "weight")
        nx.draw_networkx_edge_labels(
            mg, pos, edge_labels=labels, font_size=14, label_pos=0.3
        )

        plt.show()

    def display_with_defeat(
        self, defeat, curr_cands=None, show_undefeated=True, cmap=None
    ):
        """
        Display the margin graph with any edges that are  ``defeat`` edges highlighted in blue.

        Args:
            defeat (networkx.DiGraph): The defeat relation represented as a networkx object.
            curr_cands (List[int], optional): If set, then use the defeat relation for the profile restricted to the candidates in ``curr_cands``
            show_undefeated (bool, optional): If true, color the undefeated candidates blue and the other candidates red.
            cmap (dict, optional): The cmap used to map candidates to candidate names


        """

        cmap = cmap if cmap is not None else self.cmap
        cmap_inverse = {cname: c for c, cname in cmap.items()}
        mg_with_defeat = nx.DiGraph()

        mg = nx.DiGraph()
        mg.add_nodes_from([cmap[c] for c in self.mg.nodes])
        mg.add_edges_from([(cmap[e[0]], cmap[e[1]]) for e in self.mg.edges])

        cands = self.candidates if curr_cands is None else curr_cands

        mg_edges = list(mg.edges())
        defeat_edges = list(defeat.edges())
        edges = mg_edges + [
            (cmap[e[0]], cmap[e[1]])
            for e in defeat_edges
            if (cmap[e[0]], cmap[e[1]]) not in mg_edges
        ]

        mg_with_defeat.add_nodes_from([cmap[c] for c in cands])
        mg_with_defeat.add_edges_from(edges)

        if show_undefeated:

            undefeated_cands = [
                cmap[c1]
                for c1 in cands
                if not any([defeat.has_edge(c2, c1) for c2 in cands])
            ]
            node_colors = [
                "blue" if n in undefeated_cands else "red" for n in mg_with_defeat.nodes
            ]

        else:
            node_colors = ["#1f78b4" for n in mg_with_defeat.nodes]

        pos = nx.circular_layout(mg_with_defeat)
        nx.draw(mg, pos, width=1.5, edge_color="white")
        nx.draw_networkx_nodes(
            mg_with_defeat, pos, node_color=node_colors, node_size=700
        )
        nx.draw_networkx_labels(mg_with_defeat, pos, font_size=20, font_color="white")
        nx.draw_networkx_edges(
            mg_with_defeat,
            pos,
            edgelist=[
                (cmap[e[0]], cmap[e[1]])
                for e in defeat_edges
                if (cmap[e[0]], cmap[e[1]]) in mg_edges
            ],
            width=10,
            alpha=0.5,
            edge_color="b",
            arrowsize=25,
            min_target_margin=15,
            node_size=700,
        )
        collection = nx.draw_networkx_edges(
            mg_with_defeat,
            pos,
            edgelist=[
                (cmap[e[0]], cmap[e[1]])
                for e in defeat_edges
                if (cmap[e[0]], cmap[e[1]]) not in mg_edges
            ],
            width=10,
            alpha=0.5,
            edge_color="b",
            arrowsize=25,
            min_target_margin=15,
        )
        if collection is not None:
            for patch in collection:
                patch.set_linestyle("dashed")

        nx.draw_networkx_edges(
            mg_with_defeat,
            pos,
            edgelist=mg_edges,
            width=1.5,
            arrowsize=15,
            min_target_margin=15,
        )
        labels = {
            e: self.margin(cmap_inverse[e[0]], cmap_inverse[e[1]]) for e in mg_edges
        }

        nx.draw_networkx_edge_labels(
            mg_with_defeat, pos, edge_labels=labels, font_size=14, label_pos=0.3
        )
        ax = plt.gca()
        ax.set_frame_on(False)
        plt.show()

    def display_cycles(self, cmap=None):
        """
        Display the cycles in the margin graph.

        Args:
            cmap (dict, optional): The cmap used to map candidates to candidate names.

        """
        
        cycles = self.cycles()
        
        print(f"There {'are' if len(cycles) != 1 else 'is'} {len(cycles)} {'cycle' if len(cycles) == 1 else 'cycles'}{':' if len(cycles) > 0 else '.'} \n")
        for cycle in cycles: 
            cmap = cmap if cmap is not None else self.cmap
            mg_with_cycle = nx.DiGraph()

            mg_with_cycle.add_nodes_from([cmap[c] for c in self.candidates])
            mg_with_cycle.add_edges_from([(cmap[e[0]], cmap[e[1]]) for e in self.edges])

            cands = self.candidates
            mg_edges = list(self.edges)

            cycle_edges = [(cmap[c], cmap[cycle[cidx + 1]]) for cidx,c in enumerate(cycle[0:-1])] + [(cmap[cycle[-1]], cmap[cycle[0]])]

            cands_in_cycle = [cmap[c1] for c1 in cands if c1 in cycle]

            node_colors = ["blue" if n in cands_in_cycle else "lightgray" for n in mg_with_cycle.nodes ]

            pos = nx.circular_layout(mg_with_cycle)
            nx.draw(mg_with_cycle, pos, width=1.5, edge_color="white")
            nx.draw_networkx_nodes(
                mg_with_cycle, pos, node_color=node_colors, node_size=700
            )
            nx.draw_networkx_labels(mg_with_cycle, pos, font_size=20, font_color="white")
            nx.draw_networkx_edges(
                mg_with_cycle,
                pos,
                edgelist=cycle_edges,
                width=10,
                alpha=1.0,
                edge_color="b",
                arrowsize=25,
                min_target_margin=15,
                node_size=700,
            )

            nx.draw_networkx_edges(
                mg_with_cycle,
                pos,
                edgelist=[(cmap[e[0]], cmap[e[1]]) for e in mg_edges if (cmap[e[0]], cmap[e[1]]) not in cycle_edges],
                width=1.5,
                 edge_color="lightgray",
                arrowsize=15,
                min_target_margin=15,
            )
            labels = {
                (cmap[e[0]], cmap[e[1]]): self.margin(e[0], e[1]) for e in mg_edges
            }

            nx.draw_networkx_edge_labels(
                mg_with_cycle, pos, edge_labels=labels, font_size=14, label_pos=0.3
            )
            ax = plt.gca()
            ax.set_frame_on(False)
            plt.show()

    @classmethod
    def from_profile(cls, profile, cmap=None):
        """Generates a majority graph from a :class:`Profile`.

        :param profile: the profile
        :type profile: Profile
        :param cmap: the candidate map to use (overrides the cmap associated with this majority graph)
        :type cmap: dict[int,str], optional
        :rtype: str

        :Example:

        .. exec_code::

            from pref_voting.profiles import Profile
            from pref_voting.weighted_majority_graphs import MarginGraph
            prof = Profile([[0,1,2], [1,2,0], [2,0,1]], [2, 1, 2])
            mg = MarginGraph.from_profile(prof)
            print(mg.edges)
            print(mg.margin_matrix)

            # it is better to use the Profile method
            mg = prof.margin_graph()
            print(mg.edges)
            print(mg.margin_matrix)

        """

        cmap = profile.cmap if cmap is None else cmap
        return cls(
            profile.candidates,
            [
                (c1, c2, profile.margin(c1, c2))
                for c1 in profile.candidates
                for c2 in profile.candidates
                if profile.majority_prefers(c1, c2)
            ],
            cmap=cmap,
        )

    def __add__(self, edata): 
        """
        Return a MarginGraph in which the new margin of candidate :math:`a` over :math:`b` is the sum of the 
        existing margin of :math:`a` over :math:`b` with with the margin :math:`a` over :math:`b` in ``edata``. 
        """
        candidates = self.candidates
        new_edges = list()
        for c1, c2 in combinations(candidates, 2): 
            
            new_margin = self.margin(c1, c2) + edata.margin(c1, c2)
            
            if new_margin > 0: 
                new_edges.append((c1, c2, new_margin))
            elif new_margin < 0: 
                
                new_edges.append((c2, c1, -1 * new_margin))
            
        return MarginGraph(candidates, new_edges, cmap=self.cmap)

    def __eq__(self, other_mg): 
        """
        Return True if the margin graphs are equal (the candidates, and all edges and weights are the same); otherwise, return False
        """

        return self.candidates == other_mg.candidates and sorted(self.edges) == sorted(other_mg.edges)
    
class SupportGraph(MajorityGraph):
    """A support graph is a weighted asymmetric directed graph.  The nodes are the candidates and an edge from candidate :math:`c` to :math:`d` with weight :math:`w` means that the **support** of  :math:`c` over :math:`d` is :math:`w`.

    :param candidates: List of the candidates.  To be used as nodes in the majority graph.
    :type candidates: list[int] or  list[str]
    :param w_edges: List representing the edges in the majority graph with supports. If :math:`(c,d,(n,m))` is in the list of edges, then there is an edge from :math:`c` to :math:`d`, the support for :math:`c` over :math:`d` is :math:`n`, and the support for :math:`d` over :math:`c` is :math:`m`. 
    :type w_edges: list
    :param cmap: Dictionary mapping candidates to candidate names (strings).  If not provided, each candidate name is mapped to itself.
    :type cmap: dict[int: str], optional

    :Example:

    The following code creates a support graph in which:

    - 0 is majority preferred to 1, the number of voters who rank 0 over 1 is 4, and the number of voters who rank 1 over 0 is 3;
    - 1 is majority preferred to 2, the number of voters who rank 1 over 2 is 5, and the number of voters who rank 2 over 1 is 2; and
    - 2 is majority preferred to 0, the number of voters who rank 2 over 0 is 6, and the number of voters who rank 0 over 2 is 1.

    .. code-block:: python

        sg = SupportGraph([0, 1, 2], [(0, 1, (4, 3)), (1, 2, (5, 2)), (2, 0, (6, 1))])

    .. warning:: Currently, there is no check to that the edge relation is asymmetric.  It is assumed that the user provides an appropriate set of edges with weights.
    """

    def __init__(self, candidates, w_edges, cmap=None):
        """constructor method"""

        super().__init__(
            candidates,
            [
                (e[0], e[1]) if e[2][0] > e[2][1] else (e[1], e[0])
                for e in w_edges
                if e[2][0] != e[2][1]
            ],
            cmap=cmap,
        )

        self.s_matrix = [[0 for c2 in self.cindices] for c1 in self.cindices]
        """The support matrix, where the   :math:`(i, j)`-entry is the number of voters who rank candidate with index :math:`i` above the candidate with index :math:`j`. """

        for c1, c2, support in w_edges:
            self.s_matrix[self.cand_to_cindex(c1)][self.cand_to_cindex(c2)] = support[0]
            self.s_matrix[self.cand_to_cindex(c2)][self.cand_to_cindex(c1)] = support[1]

    @property
    def edges(self):
        """Returns a list of the weighted edges in the margin graph."""

        return [
            (c1, c2, (self.support(c1, c2), self.support(c2, c1)))
            for c1, c2 in self.mg.edges
        ]

    def margin(self, c1, c2):
        """Returns the margin of ``c1`` over ``c2``."""

        return (
            self.s_matrix[self.cand_to_cindex(c1)][self.cand_to_cindex(c2)]
            - self.s_matrix[self.cand_to_cindex(c2)][self.cand_to_cindex(c1)]
        )

    def support(self, c1, c2):
        """Returns the support of ``c1`` over ``c2``."""

        return self.s_matrix[self.cand_to_cindex(c1)][self.cand_to_cindex(c2)]

    def majority_prefers(self, c1, c2):
        """Returns True if ``c1`` is majority preferred to ``c2``."""

        return (
            self.s_matrix[self.cand_to_cindex(c1)][self.cand_to_cindex(c2)]
            > self.s_matrix[self.cand_to_cindex(c2)][self.cand_to_cindex(c1)]
        )

    def is_tied(self, c1, c2):
        """Returns True if ``c1`` is tied with  ``c2``."""

        return (
            self.s_matrix[self.cand_to_cindex(c1)][self.cand_to_cindex(c2)]
            == self.s_matrix[self.cand_to_cindex(c2)][self.cand_to_cindex(c1)]
        )

    def strength_matrix(self, curr_cands = None, strength_function = None): 
        """
        Return the strength matrix of the profile.  The strength matrix is a matrix where the entry in row :math:`i` and column :math:`j` is the number of voters that rank the candidate with index :math:`i` over the candidate with index :math:`j`.  If ``curr_cands`` is provided, then the strength matrix is restricted to the candidates in ``curr_cands``.  If ``strength_function`` is provided, then the strength matrix is computed using the strength function."""
        
        if curr_cands is not None: 
            cindices = [cidx for cidx, _ in enumerate(curr_cands)]
            cindex_to_cand = lambda cidx: curr_cands[cidx]
            cand_to_cindex = lambda c: cindices[curr_cands.index(c)]
            strength_function = self.support if strength_function is None else strength_function
            strength_matrix = np.array([[strength_function(cindex_to_cand(a_idx), cindex_to_cand(b_idx)) for b_idx in cindices] for a_idx in cindices])
        else:  
            cindices = self.cindices
            cindex_to_cand = self.cindex_to_cand
            cand_to_cindex = self.cand_to_cindex
            strength_matrix = np.array(self.s_matrix) if strength_function is None else np.array([[strength_function(cindex_to_cand(a_idx), cindex_to_cand(b_idx)) for b_idx in cindices] for a_idx in cindices])

        return strength_matrix, cand_to_cindex

    def remove_candidates(self, cands_to_ignore):
        """Remove all candidates from ``cands_to_ignore`` from the Majority Graph.

        :param cands_to_ignore: list of candidates to remove from the profile
        :type cands_to_ignore: list[int]
        :returns: a majority graph with candidates from ``cands_to_ignore`` removed and a dictionary mapping the candidates from the new profile to the original candidate names.

        :Example:

        .. exec_code::

            from pref_voting.weighted_majority_graphs import SupportGraph
            sg = SupportGraph([0, 1, 2], [(0, 1, (11, 1)), (1, 2, (5, 13)), (2, 0, (5, 10))])
            print(f"Candidates: {sg.candidates}")
            print(f"Edges: {sg.edges}")
            sg_new = sg.remove_candidates([1])
            print(f"Candidates: {sg_new.candidates}")
            print(f"Edges: {sg_new.edges}")
        """

        new_cands = [c for c in self.candidates if c not in cands_to_ignore]

        new_edges = [e for e in self.edges if e[0] in new_cands and e[1] in new_cands]

        new_cmap = {c: cname for c, cname in self.cmap.items() if c in new_cands}

        return SupportGraph(new_cands, new_edges, cmap=new_cmap)


    def display(self, curr_cands=None, cmap=None):
        """Display a support graph (restricted to ``curr_cands``) using networkx.draw.

        :param cmap: the candidate map to use (overrides the cmap associated with this majority graph)
        :type cmap: dict[int,str], optional
        :param curr_cands: list of candidates
        :type curr_cands: list[int], optional
        :rtype: None

        """

        cmap = cmap if cmap is not None else self.cmap
        curr_cands = self.candidates if curr_cands is None else curr_cands

        mg = nx.DiGraph()
        mg.add_nodes_from([cmap[c] for c in curr_cands])
        mg.add_weighted_edges_from(
            [
                (cmap[c1], cmap[c2], self.support(c1, c2))
                for c1, c2 in self.mg.edges
                if c1 in curr_cands and c2 in curr_cands
            ]
        )

        pos = nx.circular_layout(mg)

        nx.draw(
            mg,
            pos,
            font_size=20,
            font_color="white",
            node_size=700,
            width=1.5,
            with_labels=True,
        )
        labels = nx.get_edge_attributes(mg, "weight")
        nx.draw_networkx_edge_labels(
            mg, pos, edge_labels=labels, font_size=14, label_pos=0.3
        )

        plt.show()

    @classmethod
    def from_profile(cls, profile, cmap=None):
        """Generates a support graph from a :class:`Profile`.

        :param profile: the profile
        :type profile: Profile
        :param cmap: the candidate map to use (overrides the cmap associated with this majority graph)
        :type cmap: dict[int,str], optional
        :rtype: str

        :Example:

        .. exec_code::

            from pref_voting.profiles import Profile
            from pref_voting.weighted_majority_graphs import SupportGraph
            prof = Profile([[0,1,2], [1,2,0], [2,0,1]], [2, 1, 2])
            sg = SupportGraph.from_profile(prof)
            print(sg.edges)
            print(sg.s_matrix)

            # it is better to use the Profile method
            sg = prof.support_graph()
            print(sg.edges)
            print(sg.s_matrix)

        """

        cmap = profile.cmap if cmap is None else cmap
        return cls(
            profile.candidates,
            [
                (c1, c2, (profile.support(c1, c2), profile.support(c2, c1)))
                for c1 in profile.candidates
                for c2 in profile.candidates
            ],
            cmap=cmap,
        )


###
# functions to display graphs in tikz
##
def three_cand_tikz_str(g, cmap=None):
    """Returns the TikZ code to display the graph ``g`` based on 3 candidates (may be a MajorityGraph, MarginGraph or a SupportGraph)."""

    a = g.candidates[0]
    b = g.candidates[1]
    c = g.candidates[2]

    if type(g) == MarginGraph:
        w = lambda c, d: f"node[fill=white] {{${g.margin(c,d)}$}}"
    elif type(g) == SupportGraph:
        w = lambda c, d: f"node[fill=white] {{${g.support(c,d)}$}}"
    else:
        w = lambda c, d: ""

    cmap = g.cmap if cmap is None else cmap

    nodes = f"""
\\begin{{tikzpicture}}
\\node[circle,draw,minimum width=0.25in] at (0,0) (a) {{${cmap[a]}$}}; 
\\node[circle,draw,minimum width=0.25in] at (3,0) (c) {{${cmap[c]}$}}; 
\\node[circle,draw,minimum width=0.25in] at (1.5,1.5) (b) {{${cmap[b]}$}};\n"""

    if g.majority_prefers(a, b):
        ab_edge = f"\\path[->,draw,thick] (a) to {w(a,b)} (b);\n"
    elif g.majority_prefers(b, a):
        ab_edge = f"\\path[->,draw,thick] (b) to {w(b,a)} (a);\n"
    else:
        ab_edge = ""

    if g.majority_prefers(b, c):
        bc_edge = f"\\path[->,draw,thick] (b) to {w(b,c)} (c);\n"
    elif g.majority_prefers(c, b):
        bc_edge = f"\\path[->,draw,thick] (c) to {w(c,b)} (b);\n"
    else:
        bc_edge = ""

    if g.majority_prefers(a, c):
        ac_edge = f"\\path[->,draw,thick] (a) to {w(a,c)} (c);\n"
    elif g.majority_prefers(c, a):
        ac_edge = f"\\path[->,draw,thick] (c) to {w(c,a)} (a);\n"
    else:
        ac_edge = ""

    return nodes + ab_edge + bc_edge + ac_edge + "\\end{tikzpicture}"


def four_cand_tikz_str(g, cmap=None):
    """Returns the TikZ code to display the graph ``g`` based on 4 candidates (may be a MajorityGraph, MarginGraph or a SupportGraph)."""

    a = g.candidates[0]
    b = g.candidates[1]
    c = g.candidates[2]
    d = g.candidates[3]

    if type(g) == MarginGraph:
        w = lambda c, d: f"node[fill=white] {{${g.margin(c,d)}$}}"
    elif type(g) == SupportGraph:
        w = lambda c, d: f"node[fill=white] {{${g.support(c,d)}$}}"
    else:
        w = lambda c, d: ""

    cmap = g.cmap if cmap is None else cmap

    nodes = f"""
\\begin{{tikzpicture}}
\\node[circle,draw,minimum width=0.25in] at (0,0)      (a) {{${cmap[a]}$}}; 
\\node[circle,draw,minimum width=0.25in] at (3,0)      (b) {{${cmap[b]}$}}; 
\\node[circle,draw,minimum width=0.25in] at (1.5,1.5)  (c) {{${cmap[c]}$}}; 
\\node[circle,draw,minimum width=0.25in] at (1.5,-1.5) (d) {{${cmap[d]}$}};\n"""

    if g.majority_prefers(a, b):
        ab_edge = f"\\path[->,draw,thick] (a) to[pos=.7] {w(a,b)} (b);\n"
    elif g.majority_prefers(b, a):
        ab_edge = f"\\path[->,draw,thick] (b) to[pos=.7] {w(b,a)} (a);\n"
    else:
        ab_edge = ""

    if g.majority_prefers(a, c):
        ac_edge = f"\\path[->,draw,thick] (a) to {w(a,c)} (c);\n"
    elif g.majority_prefers(c, a):
        ac_edge = f"\\path[->,draw,thick] (c) to {w(c,a)} (a);\n"
    else:
        ac_edge = ""

    if g.majority_prefers(a, d):
        ad_edge = f"\\path[->,draw,thick] (a) to {w(a,d)} (d);\n"
    elif g.majority_prefers(d, a):
        ad_edge = f"\\path[->,draw,thick] (d) to {w(d,a)} (a);\n"
    else:
        ad_edge = ""

    if g.majority_prefers(b, c):
        bc_edge = f"\\path[->,draw,thick] (b) to {w(b,c)} (c);\n"
    elif g.majority_prefers(c, b):
        bc_edge = f"\\path[->,draw,thick] (c) to {w(c,b)} (b);\n"
    else:
        bc_edge = ""

    if g.majority_prefers(b, d):
        bd_edge = f"\\path[->,draw,thick] (b) to {w(b,d)} (d);\n"
    elif g.majority_prefers(d, b):
        bd_edge = f"\\path[->,draw,thick] (d) to {w(d,b)} (b);\n"
    else:
        bd_edge = ""

    if g.majority_prefers(c, d):
        cd_edge = f"\\path[->,draw,thick] (c) to[pos=.7]  {w(c,d)} (d);\n"
    elif g.majority_prefers(d, c):
        cd_edge = f"\\path[->,draw,thick] (d) to[pos=.7]  {w(d,c)} (c);\n"
    else:
        cd_edge = ""

    return (
        nodes
        + ab_edge
        + ac_edge
        + ad_edge
        + bc_edge
        + bd_edge
        + cd_edge
        + "\\end{tikzpicture}"
    )


def five_cand_tikz_str(g, cmap=None, new_cand=None):

    candidates = list(g.candidates)

    if new_cand is not None:

        e = new_cand

        cands_minus = [_c for _c in candidates if _c != new_cand]

        a = cands_minus[0]
        b = cands_minus[1]
        c = cands_minus[2]
        d = cands_minus[3]
    else:
        a = candidates[0]
        b = candidates[1]
        c = candidates[2]
        d = candidates[3]
        e = candidates[4]

    # new_cand = candidates[4]

    node_id = {a: "a", b: "b", c: "c", d: "d", e: "e"}
    print(node_id)
    if type(g) == MarginGraph:
        w = lambda c, d: f"node[fill=white] {{${g.margin(c,d)}$}}"
    elif type(g) == SupportGraph:
        w = lambda c, d: f"node[fill=white] {{${g.support(c,d)}$}}"
    else:
        w = lambda c, d: ""

    cmap = g.cmap if cmap is None else cmap

    nodes = f"""
\\begin{{tikzpicture}}
\\node[circle,draw,minimum width=0.25in] at (2,1.5)  (a) {{${cmap[a]}$}}; 
\\node[circle,draw,minimum width=0.25in] at (0,1.5)  (b) {{${cmap[b]}$}}; 
\\node[circle,draw,minimum width=0.25in] at (0,-1.5) (c) {{${cmap[c]}$}}; 
\\node[circle,draw,minimum width=0.25in] at (2,-1.5) (d) {{${cmap[d]}$}}; 
\\node[circle,draw,minimum width=0.25in] at (3.5,0)  (e) {{${cmap[e]}$}};\n"""
    edges = [(a, b), (a, d), (a, e), (b, c), (c, d), (d, e)]
    edges_with_pos = [(a, c), (b, d), (b, e), (c, e)]

    edge_tikz_str = list()
    for c1, c2 in edges:
        if g.majority_prefers(c1, c2):
            edge_tikz_str.append(
                f"\\path[->,draw,thick] ({node_id[c1]}) to {w(c1,c2)} ({node_id[c2]});\n"
            )
        elif g.majority_prefers(c2, c1):
            edge_tikz_str.append(
                f"\\path[->,draw,thick] ({node_id[c2]}) to {w(c2,c1)} ({node_id[c1]});\n"
            )
        else:
            edge_tikz_str.append("")
    for c1, c2 in edges_with_pos:
        if g.majority_prefers(c1, c2):
            edge_tikz_str.append(
                f"\\path[->,draw,thick] ({node_id[c1]}) to[pos=.7] {w(c1,c2)} ({node_id[c2]});\n"
            )
        elif g.majority_prefers(c2, c1):
            edge_tikz_str.append(
                f"\\path[->,draw,thick] ({node_id[c2]}) to[pos=.7] {w(c2,c1)} ({node_id[c1]});\n"
            )
        else:
            edge_tikz_str.append("")

    return nodes + "".join(edge_tikz_str) + "\\end{tikzpicture}"


def to_tikz_str(g, pos, cmap=None):

    node_id = {c: string.ascii_lowercase[cidx] for cidx, c in enumerate(g.candidates)}

    if type(g) == MarginGraph:
        w = lambda c, d: f"node[fill=white] {{${g.margin(c,d)}$}}"
    elif type(g) == SupportGraph:
        w = lambda c, d: f"node[fill=white] {{${g.support(c,d)}$}}"
    else:
        w = lambda c, d: ""

    cmap = g.cmap if cmap is None else cmap

    node_tikz_str = list()

    for c in g.candidates:
        node_tikz_str.append(
            f"\\node[circle,draw,minimum width=0.25in] at ({float(2.4*list(pos[c])[0])},{float(2.6*list(pos[c])[1])})  ({node_id[c]}) {{${cmap[c]}$}};\n"
        )

    edges_tikz_str = list()
    for c1 in g.candidates:
        for c2 in g.candidates:
            if g.majority_prefers(c1, c2):
                edges_tikz_str.append(
                    f"\\path[->,draw,thick] ({node_id[c1]}) to[pos=.7] {w(c1,c2)} ({node_id[c2]});\n"
                )
            elif g.majority_prefers(c2, c1):
                edges_tikz_str.append(
                    f"\\path[->,draw,thick] ({node_id[c2]}) to[pos=.7] {w(c2,c1)} ({node_id[c1]});\n"
                )
            else:
                edges_tikz_str.append("")

    return (
        "\\begin{tikzpicture}\n"
        + "".join(node_tikz_str)
        + "".join(edges_tikz_str)
        + "\\end{tikzpicture}"
    )


def maximal_elements(g):
    """return the nodes in g with no incoming arrows."""
    return [n for n in g.nodes if g.in_degree(n) == 0]


def display_mg_with_sc(edata, curr_cands=None, cmap=None):
    """
    Display the margin graph with the Split Cycle defeat relation highlighted.
    """
    from pref_voting.margin_based_methods import split_cycle_defeat

    if type(edata) == MarginGraph:
        edata.display_with_defeat(
            split_cycle_defeat(edata, curr_cands=curr_cands, cmap=cmap)
        )
    else:
        edata.display_margin_graph_with_defeat(
            split_cycle_defeat(edata, curr_cands=curr_cands, cmap=cmap)
        )


def display_graph(g, curr_cands=None, cmap=None):
    """Helper function to display a weighted directed graph."""

    cmap = cmap if cmap is not None else {n: str(n) for n in g.nodes}

    candidates = g.nodes if curr_cands is None else curr_cands

    displayed_g = nx.DiGraph()
    displayed_g.add_nodes_from([cmap[c] for c in candidates])
    displayed_g.add_weighted_edges_from(
        [(cmap[e[0]], cmap[e[1]], e[2]["weight"]) for e in g.edges(data=True)]
    )

    pos = nx.circular_layout(displayed_g)
    nx.draw(
        displayed_g,
        pos,
        font_size=20,
        node_color="blue",
        font_color="white",
        node_size=700,
        with_labels=True,
    )
    labels = nx.get_edge_attributes(displayed_g, "weight")
    nx.draw_networkx_edge_labels(
        displayed_g, pos, edge_labels=labels, font_size=14, label_pos=0.3
    )
    plt.show()
